from __future__ import annotations

import copy
import logging
import pathlib
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Tuple, cast

import pytest

import great_expectations.exceptions as ge_exceptions
import great_expectations.expectations as gxe
from great_expectations.alias_types import PathStr
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.pyspark import types as pyspark_types
from great_expectations.core.partitioners import (
    ColumnPartitioner,
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    FileNamePartitionerMonthly,
    FileNamePartitionerYearly,
)
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import (
    CSVAsset,
    DirectoryCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.delta_asset import (
    DeltaAsset,
    DirectoryDeltaAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.json_asset import (
    DirectoryJSONAsset,
    JSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.orc_asset import (
    DirectoryORCAsset,
    ORCAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.parquet_asset import (
    DirectoryParquetAsset,
    ParquetAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.text_asset import (
    DirectoryTextAsset,
    TextAsset,
)
from great_expectations.datasource.fluent.data_connector import (
    FilesystemDataConnector,
)
from great_expectations.datasource.fluent.interfaces import (
    TestConnectionError,
)
from great_expectations.datasource.fluent.spark_file_path_datasource import (
    _SparkFilePathDatasource,
)
from great_expectations.datasource.fluent.spark_filesystem_datasource import (
    SparkFilesystemDatasource,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.interfaces import BatchSlice


logger = logging.getLogger(__name__)


@pytest.fixture
def spark_filesystem_datasource(empty_data_context, test_backends) -> SparkFilesystemDatasource:
    base_directory_rel_path = pathlib.Path("..", "..", "test_sets", "taxi_yellow_tripdata_samples")
    base_directory_abs_path = (
        pathlib.Path(__file__).parent.joinpath(base_directory_rel_path).resolve(strict=True)
    )
    spark_filesystem_datasource = SparkFilesystemDatasource(
        name="spark_filesystem_datasource",
        base_directory=base_directory_abs_path,
    )
    spark_filesystem_datasource._data_context = empty_data_context

    # Verify test directory has files we expect
    years = ["2018", "2019", "2020"]
    file_name: PathStr
    all_files: List[str] = [
        file_name.stem
        for file_name in list(pathlib.Path(spark_filesystem_datasource.base_directory).iterdir())
    ]
    # assert there are 12 files for each year
    for year in years:
        files_for_year = [
            file_name
            for file_name in all_files
            if file_name.find(f"yellow_tripdata_sample_{year}") == 0
        ]
        assert len(files_for_year) == 12

    return spark_filesystem_datasource


@pytest.mark.spark
def test_add_csv_asset_to_datasource(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    assert asset.name == "csv_asset"


if pyspark_types:
    spark_schema = pyspark_types.StructType(
        [pyspark_types.StructField("f1", pyspark_types.StringType(), True)]
    )
    additional_params = {"spark_schema": spark_schema}
else:
    additional_params = {}

add_csv_asset_all_params = {
    "sep": "sep",
    "encoding": "encoding",
    "quote": "quote",
    "escape": "escape",
    "comment": "comment",
    "header": "header",
    "infer_schema": "infer_schema",
    "ignore_leading_white_space": "ignore_leading_white_space",
    "ignore_trailing_white_space": "ignore_trailing_white_space",
    "null_value": "null_value",
    "nan_value": "nan_value",
    "positive_inf": "positive_inf",
    "negative_inf": "negative_inf",
    "date_format": "date_format",
    "timestamp_format": "timestamp_format",
    "max_columns": "max_columns",
    "max_chars_per_column": "max_chars_per_column",
    "max_malformed_log_per_partition": "max_malformed_log_per_partition",
    "mode": "PERMISSIVE",
    "column_name_of_corrupt_record": "column_name_of_corrupt_record",
    "multi_line": "multi_line",
    "char_to_escape_quote_escaping": "char_to_escape_quote_escaping",
    "sampling_ratio": "sampling_ratio",
    "enforce_schema": "enforce_schema",
    "empty_value": "empty_value",
    "locale": "locale",
    "line_sep": "line_sep",
    "path_glob_filter": "path_glob_filter",
    "recursive_file_lookup": "recursive_file_lookup",
    "modified_before": "modified_before",
    "modified_after": "modified_after",
    "unescaped_quote_handling": "STOP_AT_CLOSING_QUOTE",
}

add_csv_asset = [
    pytest.param(
        "add_csv_asset",
        {},
        id="csv_min_params",
    ),
    pytest.param(
        "add_csv_asset",
        {
            **add_csv_asset_all_params,
            **additional_params,
        },
        id="csv_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_csv_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="csv_fail_extra_params",
    ),
]

add_parquet_asset_all_params = {
    "merge_schema": "merge_schema",
    "datetime_rebase_mode": "EXCEPTION",
    "int_96_rebase_mode": "EXCEPTION",
    "path_glob_filter": "path_glob_filter",
    "recursive_file_lookup": "recursive_file_lookup",
    "modified_before": "modified_before",
    "modified_after": "modified_after",
}

add_parquet_asset = [
    pytest.param(
        "add_parquet_asset",
        {},
        id="parquet_min_params",
    ),
    pytest.param(
        "add_parquet_asset",
        add_parquet_asset_all_params,
        id="parquet_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_parquet_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="parquet_fail_extra_params",
    ),
]

add_orc_asset_all_params = {
    "merge_schema": "merge_schema",
    "path_glob_filter": "path_glob_filter",
    "recursive_file_lookup": "recursive_file_lookup",
    "modified_before": "modified_before",
    "modified_after": "modified_after",
}

add_orc_asset = [
    pytest.param(
        "add_orc_asset",
        {},
        id="orc_min_params",
    ),
    pytest.param(
        "add_orc_asset",
        add_orc_asset_all_params,
        id="orc_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_orc_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="orc_fail_extra_params",
    ),
]

add_json_asset_all_params = {
    "primitives_as_string": "primitives_as_string",
    "prefers_decimal": "prefers_decimal",
    "allow_comments": "allow_comments",
    "allow_unquoted_field_names": "allow_unquoted_field_names",
    "allow_single_quotes": "allow_single_quotes",
    "allow_numeric_leading_zero": "allow_numeric_leading_zero",
    "allow_backslash_escaping_any_character": "allow_backslash_escaping_any_character",
    "mode": "PERMISSIVE",
    "column_name_of_corrupt_record": "column_name_of_corrupt_record",
    "date_format": "date_format",
    "timestamp_format": "timestamp_format",
    "multi_line": "multi_line",
    "allow_unquoted_control_chars": "allow_unquoted_control_chars",
    "line_sep": "line_sep",
    "sampling_ratio": "sampling_ratio",
    "drop_field_if_all_null": "drop_field_if_all_null",
    "encoding": "encoding",
    "locale": "locale",
    "path_glob_filter": "path_glob_filter",
    "recursive_file_lookup": "recursive_file_lookup",
    "modified_before": "modified_before",
    "modified_after": "modified_after",
    "allow_non_numeric_numbers": "allow_non_numeric_numbers",
}

add_json_asset = [
    pytest.param(
        "add_json_asset",
        {},
        id="json_min_params",
    ),
    pytest.param(
        "add_json_asset",
        {
            **add_json_asset_all_params,
            **additional_params,
        },
        id="json_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_json_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="json_fail_extra_params",
    ),
]

add_text_asset_all_params = {
    "wholetext": True,
    "line_sep": "line_sep",
    "path_glob_filter": "path_glob_filter",
    "recursive_file_lookup": "recursive_file_lookup",
    "modified_before": "modified_before",
    "modified_after": "modified_after",
}

add_text_asset = [
    pytest.param(
        "add_text_asset",
        {},
        id="text_min_params",
    ),
    pytest.param(
        "add_text_asset",
        add_text_asset_all_params,
        id="text_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_text_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="text_fail_extra_params",
    ),
]

add_delta_asset_all_params = {
    "timestamp_as_of": "timestamp_as_of",
    "version_as_of": "version_as_of",
}

add_delta_asset = [
    pytest.param(
        "add_delta_asset",
        {},
        id="delta_min_params",
    ),
    pytest.param(
        "add_delta_asset",
        add_delta_asset_all_params,
        id="delta_all_params_20230512",
    ),
    pytest.param(
        "add_delta_asset",
        {
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="delta_fail_extra_params",
    ),
]

add_asset_test_params = []
add_asset_test_params += add_csv_asset
add_asset_test_params += add_parquet_asset
add_asset_test_params += add_orc_asset
add_asset_test_params += add_json_asset
add_asset_test_params += add_text_asset
add_asset_test_params += add_delta_asset


@pytest.mark.unit
@pytest.mark.parametrize(
    "add_method_name,add_method_params",
    add_asset_test_params,
)
def test_add_asset_with_asset_specific_params(
    spark_filesystem_datasource: SparkFilesystemDatasource,
    add_method_name: str,
    add_method_params: dict,
):
    asset = getattr(spark_filesystem_datasource, add_method_name)(
        name="asset_name", **add_method_params
    )
    assert asset.name == "asset_name"
    for param, value in add_method_params.items():
        if param == "spark_schema":
            struct_type = cast("pyspark_types.StructType", value)
            assert getattr(asset, param) == struct_type.jsonValue()
        else:
            assert getattr(asset, param) == value


_SPARK_ASSET_TYPES = [
    (CSVAsset, {"name": "asset_name"}),
    (DirectoryCSVAsset, {"name": "asset_name", "data_directory": "data_directory"}),
    (ParquetAsset, {"name": "asset_name"}),
    (DirectoryParquetAsset, {"name": "asset_name", "data_directory": "data_directory"}),
    (ORCAsset, {"name": "asset_name"}),
    (DirectoryORCAsset, {"name": "asset_name", "data_directory": "data_directory"}),
    (JSONAsset, {"name": "asset_name"}),
    (DirectoryJSONAsset, {"name": "asset_name", "data_directory": "data_directory"}),
    (TextAsset, {"name": "asset_name"}),
    (DirectoryTextAsset, {"name": "asset_name", "data_directory": "data_directory"}),
    (DeltaAsset, {"name": "asset_name"}),
    (DirectoryDeltaAsset, {"name": "asset_name", "data_directory": "data_directory"}),
]


@pytest.mark.unit
def test_all_spark_file_path_asset_types_tested():
    """Make sure all the available assets are contained in the fixture used for other tests."""
    asset_types_in_fixture = {a[0] for a in _SPARK_ASSET_TYPES}
    assert asset_types_in_fixture == set(_SparkFilePathDatasource.asset_types)


@pytest.mark.unit
@pytest.mark.parametrize(
    "asset_type,required_fields",
    [
        pytest.param(
            asset_type,
            required_fields,
            id=asset_type.__name__,
        )
        for (asset_type, required_fields) in _SPARK_ASSET_TYPES
    ],
)
def test__get_reader_options_include(asset_type: PathDataAsset, required_fields: dict):
    """Make sure options are in fields."""
    fields = set(asset_type.__fields__.keys())
    asset = asset_type.validate(required_fields)
    for option in asset._get_reader_options_include():
        assert option in fields


add_directory_csv_asset = [
    pytest.param(
        "add_directory_csv_asset",
        {"data_directory": "some_directory"},
        id="directory_csv_min_params",
    ),
    pytest.param(
        "add_directory_csv_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_csv_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_csv_asset",
        {
            **add_csv_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
            **additional_params,
        },
        id="directory_csv_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_csv_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_csv_fail_extra_params",
    ),
]

add_directory_parquet_asset = [
    pytest.param(
        "add_directory_parquet_asset",
        {"data_directory": "some_directory"},
        id="directory_parquet_min_params",
    ),
    pytest.param(
        "add_directory_parquet_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_parquet_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_parquet_asset",
        {
            **add_parquet_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
        },
        id="directory_parquet_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_parquet_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_parquet_fail_extra_params",
    ),
]


add_directory_orc_asset = [
    pytest.param(
        "add_directory_orc_asset",
        {"data_directory": "some_directory"},
        id="directory_orc_min_params",
    ),
    pytest.param(
        "add_directory_orc_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_orc_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_orc_asset",
        {
            **add_orc_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
        },
        id="directory_orc_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_orc_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_orc_fail_extra_params",
    ),
]


add_directory_json_asset = [
    pytest.param(
        "add_directory_json_asset",
        {"data_directory": "some_directory"},
        id="directory_json_min_params",
    ),
    pytest.param(
        "add_directory_json_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_json_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_json_asset",
        {
            **add_json_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
            **additional_params,
        },
        id="directory_json_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_json_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_json_fail_extra_params",
    ),
]


add_directory_text_asset = [
    pytest.param(
        "add_directory_text_asset",
        {"data_directory": "some_directory"},
        id="directory_text_min_params",
    ),
    pytest.param(
        "add_directory_text_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_text_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_text_asset",
        {
            **add_text_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
        },
        id="directory_text_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_text_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_text_fail_extra_params",
    ),
]


add_directory_delta_asset = [
    pytest.param(
        "add_directory_delta_asset",
        {"data_directory": "some_directory"},
        id="directory_delta_min_params",
    ),
    pytest.param(
        "add_directory_delta_asset",
        {"data_directory": pathlib.Path("some_directory")},
        id="directory_delta_min_params_pathlib",
    ),
    pytest.param(
        "add_directory_delta_asset",
        {
            **add_delta_asset_all_params,
            **{
                "data_directory": "some_directory",
            },
        },
        id="directory_delta_all_params_pyspark_3_4_0",
    ),
    pytest.param(
        "add_directory_delta_asset",
        {
            "data_directory": "some_directory",
            "this_param_does_not_exist": "param_does_not_exist",
        },
        marks=pytest.mark.xfail(
            reason="param_does_not_exist",
            strict=True,
            raises=pydantic.ValidationError,
        ),
        id="directory_delta_fail_extra_params",
    ),
]

add_directory_asset_test_params = []
add_directory_asset_test_params += add_directory_csv_asset
add_directory_asset_test_params += add_directory_parquet_asset
add_directory_asset_test_params += add_directory_orc_asset
add_directory_asset_test_params += add_directory_json_asset
add_directory_asset_test_params += add_directory_text_asset
add_directory_asset_test_params += add_directory_delta_asset


@pytest.mark.unit
@pytest.mark.parametrize(
    "add_method_name,add_method_params",
    add_directory_asset_test_params,
)
def test_add_directory_asset_with_asset_specific_params(
    spark_filesystem_datasource: SparkFilesystemDatasource,
    add_method_name: str,
    add_method_params: dict,
):
    asset = getattr(spark_filesystem_datasource, add_method_name)(
        name="asset_name", **add_method_params
    )
    assert asset.name == "asset_name"
    for param, value in add_method_params.items():
        if param == "data_directory":
            assert getattr(asset, param) == pathlib.Path(str(value))
        elif param == "spark_schema":
            struct_type = cast("pyspark_types.StructType", value)
            assert getattr(asset, param) == struct_type.jsonValue()
        else:
            assert getattr(asset, param) == value


@pytest.mark.unit
def test_add_csv_asset_with_batching_regex_to_datasource(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
def test_construct_csv_asset_directly():
    asset = CSVAsset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
def test_csv_asset_with_batching_regex_named_parameters(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    batching_regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=batching_regex)
    options = asset.get_batch_parameters_keys(partitioner=batch_def.partitioner)
    assert options == ("path", "year", "month")


@pytest.mark.unit
def test_csv_asset_with_non_string_batching_regex_named_parameters(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    with pytest.raises(ge_exceptions.InvalidBatchRequestError):
        # year is an int which will raise an error
        asset.build_batch_request({"year": 2018, "month": "04"})


@pytest.mark.spark
@pytest.mark.parametrize(
    "path",
    [
        pytest.param("samples_2020", id="str"),
        pytest.param(pathlib.Path("samples_2020"), id="pathlib.Path"),
    ],
)
def test_get_batch_identifiers_list_from_directory_one_batch(
    path: PathStr,
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    """What does this test and why?

    A "directory" asset should only have a single batch."""
    asset = spark_filesystem_datasource.add_directory_csv_asset(
        name="csv_asset",
        data_directory=path,
        header=True,
        infer_schema=True,
    )
    batch_def = asset.add_batch_definition_whole_directory(name="test batch def")
    request = batch_def.build_batch_request()
    batches = asset.get_batch_identifiers_list(request)
    assert len(batches) == 1


@pytest.mark.spark
@pytest.mark.parametrize(
    "path",
    [
        pytest.param("samples_2020", id="str"),
        pytest.param(pathlib.Path("samples_2020"), id="pathlib.Path"),
    ],
)
def test_get_batch_list_from_directory_merges_files(
    path: PathStr,
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    """What does this test and why?

    Adding a "directory" asset should only add a single batch merging all files into one dataframe.

    Marked as an integration test since this uses the execution engine to actually load the files.
    """
    asset = spark_filesystem_datasource.add_directory_csv_asset(
        name="csv_asset",
        data_directory=path,
        header=True,
        infer_schema=True,
    )
    batch_def = asset.add_batch_definition_whole_directory("test batch def")
    request = batch_def.build_batch_request()
    batch = asset.get_batch(request)
    batch_data = batch.data
    # The directory contains 12 files with 10,000 records each so the batch data
    # (spark dataframe) should contain 120,000 records:
    assert batch_data.dataframe.count() == 12 * 10000  # type: ignore[attr-defined] # FIXME CoP


@pytest.mark.spark
def test_get_batch_list_from_fully_specified_batch_request(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    batch_def = asset.add_batch_definition_monthly(
        name="Fully Specified Batch Test",
        regex=r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv",
    )
    batch = batch_def.get_batch(batch_parameters={"year": "2018", "month": "04"})
    assert batch.batch_request.datasource_name == spark_filesystem_datasource.name
    assert batch.batch_request.data_asset_name == asset.name

    path = "yellow_tripdata_sample_2018-04.csv"
    assert batch.batch_request.options == {"path": path, "year": "2018", "month": "04"}
    assert batch.metadata == {"path": path, "year": "2018", "month": "04"}

    assert batch.id == "spark_filesystem_datasource-csv_asset-year_2018-month_04"


@pytest.mark.spark
def test_get_batch_list_from_partially_specified_batch_request(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    # Verify test directory has files that don't match what we will query for
    file_name: PathStr
    all_files: List[str] = [
        file_name.stem
        for file_name in list(pathlib.Path(spark_filesystem_datasource.base_directory).iterdir())
    ]
    # assert there are files that are not csv files
    assert any(not file_name.endswith("csv") for file_name in all_files)
    # assert there are 12 files from 2018
    files_for_2018 = [file_name for file_name in all_files if file_name.find("2018") >= 0]
    assert len(files_for_2018) == 12

    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    batch_def = asset.add_batch_definition_monthly(
        name="Partially Specified Batch Test",
        regex=r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv",
    )

    request = batch_def.build_batch_request(batch_parameters={"year": "2018"})

    batch_identifiers_list = asset.get_batch_identifiers_list(request)
    assert (len(batch_identifiers_list)) == 12
    batch_filenames = [pathlib.Path(batch["path"]).stem for batch in batch_identifiers_list]
    assert set(files_for_2018) == set(batch_filenames)

    @dataclass(frozen=True)
    class YearMonth:
        year: str
        month: str

    expected_year_month = {YearMonth(year="2018", month=format(m, "02d")) for m in range(1, 13)}
    batch_year_month = {
        YearMonth(year=batch["year"], month=batch["month"]) for batch in batch_identifiers_list
    }
    assert expected_year_month == batch_year_month


@pytest.mark.spark
@pytest.mark.parametrize(
    "sort_ascending",
    [pytest.param(True), pytest.param(False)],
)
def test_spark_sorter(spark_filesystem_datasource: SparkFilesystemDatasource, sort_ascending: bool):
    # arrange
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    regex = re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
    batch_def = asset.add_batch_definition_monthly(
        name="test-batch-def", regex=regex, sort_ascending=sort_ascending
    )
    batch_request = batch_def.build_batch_request()

    # act
    batches = asset.get_batch_identifiers_list(batch_request)

    # assert
    assert (len(batches)) == 36

    reverse = sort_ascending is False
    sorted_batches = sorted(
        batches,
        key=lambda batch: (batch.get("year"), batch.get("month")),
        reverse=reverse,
    )
    assert sorted_batches == batches


@pytest.mark.spark
@pytest.mark.parametrize(
    "batch_slice, expected_batch_count",
    [
        ("[-3:]", 3),
        ("[5:9]", 4),
        ("[:10:2]", 5),
        (slice(-3, None), 3),
        (slice(5, 9), 4),
        (slice(0, 10, 2), 5),
        ("-5", 1),
        ("-1", 1),
        (11, 1),
        (0, 1),
        ([3], 1),
        (None, 12),
        ("", 12),
    ],
)
def test_spark_slice_batch_count(
    spark_filesystem_datasource: SparkFilesystemDatasource,
    batch_slice: BatchSlice,
    expected_batch_count: int,
) -> None:
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    batching_regex = re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
    batch_request = asset.build_batch_request(
        options={"year": "2019"},
        batch_slice=batch_slice,
        partitioner=FileNamePartitionerMonthly(regex=batching_regex),
    )
    batches = asset.get_batch_identifiers_list(batch_request=batch_request)
    assert len(batches) == expected_batch_count


def bad_batching_regex_config(
    csv_path: pathlib.Path,
) -> tuple[re.Pattern, TestConnectionError]:
    batching_regex = re.compile(r"green_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
    test_connection_error = TestConnectionError(
        f"""No file at base_directory path "{csv_path.resolve()}" matched regular expressions pattern "{batching_regex.pattern}" and/or glob_directive "**/*" for DataAsset "csv_asset"."""  # noqa: E501 # FIXME CoP
    )
    return batching_regex, test_connection_error


@pytest.fixture(params=[bad_batching_regex_config])
def datasource_test_connection_error_messages(
    csv_path: pathlib.Path,
    spark_filesystem_datasource: SparkFilesystemDatasource,
    request,
) -> tuple[SparkFilesystemDatasource, TestConnectionError]:
    _, test_connection_error = request.param(csv_path=csv_path)
    csv_asset = CSVAsset(  # type: ignore[call-arg] # missing args
        name="csv_asset",
    )
    csv_asset._datasource = spark_filesystem_datasource
    spark_filesystem_datasource.assets = [
        csv_asset,
    ]
    csv_asset._data_connector = FilesystemDataConnector(
        datasource_name=spark_filesystem_datasource.name,
        data_asset_name=csv_asset.name,
        base_directory=spark_filesystem_datasource.base_directory,
        data_context_root_directory=spark_filesystem_datasource.data_context_root_directory,
    )
    csv_asset._test_connection_error_message = test_connection_error
    return spark_filesystem_datasource, test_connection_error


@pytest.mark.spark
def test_get_batch_identifiers_list_does_not_modify_input_batch_request(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="Test Batch Definition", regex=regex)
    request = batch_def.build_batch_request({"year": "2018"})
    request_before_call = copy.deepcopy(request)
    batches = asset.get_batch_identifiers_list(request)
    # We assert the request before the call to get_batch_identifiers_list is equal to the request after the  # noqa: E501 # FIXME CoP
    # call. This test exists because this call was modifying the request.
    assert request == request_before_call
    # We get all 12 batches, one for each month of 2018.
    assert len(batches) == 12


@pytest.mark.spark
def test_get_batch_does_not_modify_input_batch_request(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
    )
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="Test Batch Definition", regex=regex)
    request = batch_def.build_batch_request({"year": "2018"})
    request_before_call = copy.deepcopy(request)
    _ = asset.get_batch(request)
    # We assert the request before the call to get_batch is equal to the request after the  # noqa: E501 # FIXME CoP
    # call. This test exists because this call was modifying the request.
    assert request == request_before_call


@pytest.mark.spark
def test_add_csv_asset_with_batch_metadata(
    spark_filesystem_datasource: SparkFilesystemDatasource,
):
    asset_specified_metadata = {"asset_level_metadata": "my_metadata"}
    asset = spark_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        header=True,
        infer_schema=True,
        batch_metadata=asset_specified_metadata,
    )
    batch_def = asset.add_batch_definition_monthly(
        name="Test Batch Metadata",
        regex=r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv",
    )
    batch_parameters = {"year": "2018", "month": "05"}

    request = batch_def.build_batch_request(batch_parameters)
    batch = asset.get_batch(request)
    assert batch.metadata == {
        "path": "yellow_tripdata_sample_2018-05.csv",
        **batch_parameters,
        **asset_specified_metadata,
    }


@pytest.fixture
def directory_asset_with_no_partitioner(
    spark_filesystem_datasource: SparkFilesystemDatasource,
) -> DirectoryCSVAsset:
    asset = spark_filesystem_datasource.add_directory_csv_asset(
        name="directory_csv_asset_no_partitioner",
        data_directory="first_ten_trips_in_each_file",
        header=True,
        infer_schema=True,
    )
    return asset


@pytest.fixture
def expected_num_records_directory_asset_no_partitioner_2020_passenger_count_2(
    directory_asset_with_no_partitioner: DirectoryCSVAsset,
) -> int:
    pre_partitioner_batch = directory_asset_with_no_partitioner.get_batch(
        directory_asset_with_no_partitioner.build_batch_request()
    )
    pre_partitioner_batch_data = pre_partitioner_batch.data
    expected_num_records = pre_partitioner_batch_data.dataframe.filter(  # type: ignore[attr-defined] # FIXME CoP
        F.col("pickup_datetime").contains("2018-01-11")
    ).count()
    assert expected_num_records == 3, "Check that the referenced data hasn't changed"
    return expected_num_records


@pytest.fixture
def directory_asset(
    spark_filesystem_datasource: SparkFilesystemDatasource,
) -> DirectoryCSVAsset:
    asset = spark_filesystem_datasource.add_directory_csv_asset(
        name="directory_csv_asset_with_partitioner",
        data_directory="first_ten_trips_in_each_file",
        header=True,
        infer_schema=True,
    )
    return asset


@pytest.fixture
def daily_partitioner():
    return ColumnPartitionerDaily(column_name="pickup_datetime")


@pytest.fixture
def daily_batch_parameters_and_expected_result() -> Tuple[Dict[str, int], int]:
    batch_parameters = {"year": 2018, "month": 1, "day": 11}
    expected_result = 3
    return batch_parameters, expected_result


@pytest.fixture
def monthly_batch_parameters_and_expected_result() -> Tuple[Dict[str, int], int]:
    batch_parameters = {"year": 2018, "month": 1}
    expected_result = 10
    return batch_parameters, expected_result


@pytest.fixture
def yearly_batch_parameters_and_expected_result() -> Tuple[Dict[str, int], int]:
    batch_parameters = {"year": 2018}
    expected_result = 120
    return batch_parameters, expected_result


@pytest.fixture
def whole_directory_batch_parameters_and_expected_result() -> Tuple[Dict[str, int], int]:
    batch_parameters: Dict[str, int] = {}
    expected_result = 360
    return batch_parameters, expected_result


class TestPartitionerDirectoryAsset:
    @pytest.mark.spark
    def test_daily_batch_definition_workflow(
        self,
        directory_asset: DirectoryCSVAsset,
    ):
        batch_parameters = {"year": 2018, "month": 1, "day": 11}
        expected_result = 3
        batch_def = directory_asset.add_batch_definition_daily(
            name="daily", column="pickup_datetime"
        )
        batch = batch_def.get_batch(batch_parameters=batch_parameters)
        assert batch.validate(gxe.ExpectTableRowCountToEqual(value=expected_result)).success

    @pytest.mark.spark
    def test_monthly_batch_definition_workflow(
        self,
        directory_asset: DirectoryCSVAsset,
    ):
        batch_parameters = {"year": 2018, "month": 1}
        expected_result = 10
        batch_def = directory_asset.add_batch_definition_monthly(
            name="monthly", column="pickup_datetime"
        )
        batch = batch_def.get_batch(batch_parameters=batch_parameters)
        assert batch.validate(gxe.ExpectTableRowCountToEqual(value=expected_result)).success

    @pytest.mark.spark
    def test_yearly_batch_definition_workflow(
        self,
        directory_asset: DirectoryCSVAsset,
        yearly_batch_parameters_and_expected_result: Tuple[Dict[str, int], int],
    ):
        batch_parameters = {"year": 2018}
        expected_result = 120
        batch_def = directory_asset.add_batch_definition_yearly(
            name="yearly", column="pickup_datetime"
        )
        batch = batch_def.get_batch(batch_parameters=batch_parameters)
        assert batch.validate(gxe.ExpectTableRowCountToEqual(value=expected_result)).success

    @pytest.mark.spark
    def test_whole_table_batch_definition_workflow(
        self,
        directory_asset: DirectoryCSVAsset,
        whole_directory_batch_parameters_and_expected_result: Tuple[Dict[str, int], int],
    ):
        batch_parameters: Dict[str, int] = {}
        expected_result = 360
        batch_def = directory_asset.add_batch_definition_whole_directory(name="whole directory")
        batch = batch_def.get_batch(batch_parameters=batch_parameters)
        assert batch.validate(gxe.ExpectTableRowCountToEqual(value=expected_result)).success

    @pytest.mark.spark
    @pytest.mark.parametrize(
        "partitioner,expected_keys",
        [
            pytest.param(
                ColumnPartitionerDaily(column_name="foo"),
                ("path", "year", "month", "day"),
                id="Daily Partitioner",
            ),
            pytest.param(
                ColumnPartitionerMonthly(column_name="foo"),
                (
                    "path",
                    "year",
                    "month",
                ),
                id="Monthly Partitioner",
            ),
            pytest.param(
                ColumnPartitionerYearly(column_name="foo"),
                (
                    "path",
                    "year",
                ),
                id="Yearly Partitioner",
            ),
            pytest.param(None, ("path",), id="No Partitioner"),
        ],
    )
    def test_get_batch_parameters_keys_with_partitioner(
        self,
        directory_asset: DirectoryCSVAsset,
        partitioner: ColumnPartitioner,
        expected_keys: Tuple[str, ...],
    ):
        assert directory_asset.get_batch_parameters_keys(partitioner=partitioner) == expected_keys


@pytest.fixture
def file_asset_with_no_partitioner(
    spark_filesystem_datasource: SparkFilesystemDatasource,
) -> CSVAsset:
    asset = spark_filesystem_datasource.add_csv_asset(
        name="file_csv_asset_no_partitioner",
        header=True,
        infer_schema=True,
    )
    return asset


@pytest.fixture
def expected_num_records_file_asset_no_partitioner_2020_10_passenger_count_2(
    file_asset_with_no_partitioner: CSVAsset,
) -> int:
    single_batch_batch_request = file_asset_with_no_partitioner.build_batch_request(
        {"year": "2020", "month": "11"}
    )
    batch = file_asset_with_no_partitioner.get_batch(single_batch_batch_request)
    pre_partitioner_batch_data = batch.data
    expected_num_records = pre_partitioner_batch_data.dataframe.filter(  # type: ignore[attr-defined] # FIXME CoP
        F.col("passenger_count") == 2
    ).count()
    assert expected_num_records == 2, "Check that the referenced data hasn't changed"
    return expected_num_records


@pytest.fixture
def expected_num_records_file_asset_no_partitioner_2020_10(
    file_asset_with_no_partitioner: CSVAsset,
) -> int:
    single_batch_batch_request = file_asset_with_no_partitioner.build_batch_request(
        {"year": "2020", "month": "11"},
        partitioner=FileNamePartitionerMonthly(
            regex=re.compile(
                r"first_ten_trips_in_each_file/yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
            )
        ),
    )
    batch = file_asset_with_no_partitioner.get_batch(single_batch_batch_request)

    pre_partitioner_batch_data = batch.data

    expected_num_records = (
        pre_partitioner_batch_data.dataframe.filter(  # type: ignore[attr-defined] # FIXME CoP
            F.year(F.col("pickup_datetime")) == 2020
        )
        .filter(F.month(F.col("pickup_datetime")) == 11)
        .count()
    )
    assert expected_num_records == 10, "Check that the referenced data hasn't changed"
    return expected_num_records


@pytest.fixture
def file_asset(
    spark_filesystem_datasource: SparkFilesystemDatasource,
) -> CSVAsset:
    asset = spark_filesystem_datasource.add_csv_asset(
        name="file_csv_asset_with_partitioner",
        header=True,
        infer_schema=True,
    )
    return asset


class TestPartitionerFileAsset:
    @pytest.mark.spark
    @pytest.mark.xfail(strict=True, reason="Will fix or refactor as part of V1-306")
    def test_parameter_keys_with_partitioner_file_asset_batch_parameters(
        self, file_asset, daily_partitioner
    ):
        assert file_asset.get_batch_parameters_keys(partitioner=daily_partitioner) == (
            "year",
            "month",
            "path",
            "passenger_count",
        )

    @pytest.mark.spark
    @pytest.mark.xfail(strict=True, reason="Will fix or refactor as part of V1-306")
    def test_get_batch_with_partitioner_file_asset_one_batch_size(
        self,
        file_asset,
        daily_partitioner,
        expected_num_records_file_asset_no_partitioner_2020_10_passenger_count_2: int,
    ):
        post_partitioner_batch_request = file_asset.build_batch_request(
            options={"year": "2020", "month": "11", "passenger_count": 2},
            partitioner=daily_partitioner,
        )
        post_partitioner_batch_list = file_asset.get_batch(post_partitioner_batch_request)

        # Make sure we only have passenger_count == 2 in our batch data
        post_partitioner_batch_data = post_partitioner_batch_list.data

        assert (
            post_partitioner_batch_data.dataframe.filter(F.col("passenger_count") == 2).count()
            == expected_num_records_file_asset_no_partitioner_2020_10_passenger_count_2
        )
        assert (
            post_partitioner_batch_data.dataframe.filter(F.col("passenger_count") != 2).count() == 0
        )

    @pytest.mark.spark
    @pytest.mark.xfail(strict=True, reason="Will fix or refactor as part of V1-306")
    def test_add_file_csv_asset_with_partitioner_conflicting_identifier_batch_parameters(
        self, file_asset_with_no_partitioner: CSVAsset
    ):
        regex = re.compile(
            r"first_ten_trips_in_each_file/yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
        )
        asset_with_conflicting_partitioner = file_asset_with_no_partitioner
        partitioner = FileNamePartitionerYearly(regex=regex)
        assert asset_with_conflicting_partitioner.get_batch_parameters_keys(
            partitioner=partitioner
        ) == (
            "year",
            "month",
            "path",
        )

    @pytest.mark.spark
    def test_add_file_csv_asset_with_partitioner_conflicting_identifier_gets_a_batch(
        self, file_asset_with_no_partitioner: CSVAsset
    ):
        regex = re.compile(
            r"first_ten_trips_in_each_file/yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
        )

        asset = file_asset_with_no_partitioner
        partitioner = FileNamePartitionerMonthly(regex=regex)

        post_partitioner_batch_request = asset.build_batch_request(
            options={"year": "2020", "month": "11"}, partitioner=partitioner
        )
        asset.get_batch(post_partitioner_batch_request)
        # no errors!

    @pytest.mark.spark
    def test_add_file_csv_asset_with_partitioner_conflicting_identifier_gets_correct_data(
        self,
        file_asset_with_no_partitioner: CSVAsset,
        expected_num_records_file_asset_no_partitioner_2020_10: int,
    ):
        regex = re.compile(
            r"first_ten_trips_in_each_file/yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
        )
        asset = file_asset_with_no_partitioner
        partitioner = FileNamePartitionerMonthly(regex=regex)

        post_partitioner_batch_request = asset.build_batch_request(
            options={"year": "2020", "month": "11"}, partitioner=partitioner
        )
        post_partitioner_batch = asset.get_batch(post_partitioner_batch_request)
        post_partitioner_batch_data = post_partitioner_batch.data

        assert (
            post_partitioner_batch_data.dataframe.count()  # type: ignore[attr-defined] # FIXME CoP
            == expected_num_records_file_asset_no_partitioner_2020_10
        )



from __future__ import annotations

import copy
import logging
import warnings
from datetime import date, datetime
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Final,
    Generic,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from typing_extensions import Annotated, Never

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.batch_spec import (
    BatchSpec,
    RuntimeQueryBatchSpec,
    SqlAlchemyDatasourceBatchSpec,
)
from great_expectations.core.partitioners import (
    ColumnPartitioner,
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    PartitionerColumnValue,
    PartitionerConvertedDatetime,
    PartitionerDatetimePart,
    PartitionerDividedInteger,
    PartitionerModInteger,
    PartitionerMultiColumnValue,
)
from great_expectations.datasource.fluent.batch_request import (
    BatchRequest,
)
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
    _check_config_substitutions_needed,
)
from great_expectations.datasource.fluent.constants import _DATA_CONNECTOR_NAME
from great_expectations.datasource.fluent.fluent_base_model import FluentBaseModel
from great_expectations.datasource.fluent.interfaces import (
    Batch,
    DataAsset,
    Datasource,
    DatasourceT,
    GxDatasourceWarning,
    PartitionerProtocol,
    TestConnectionError,
)
from great_expectations.exceptions.exceptions import (
    NoAvailableBatchesError,
    SqlAddBatchDefinitionError,
)
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.partition_and_sample.data_partitioner import (
    DatePart,
)
from great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_partitioner import (
    SqlAlchemyDataPartitioner,
)

if TYPE_CHECKING:
    from sqlalchemy.sql import quoted_name  # noqa: TID251 # type-checking only

    # We re-import sqlalchemy here to make type-checking and our compatability layer
    # play nice with one another
    from great_expectations.compatibility import sqlalchemy
    from great_expectations.core.batch_definition import BatchDefinition
    from great_expectations.datasource.fluent import BatchParameters
    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
        BatchSlice,
    )

LOGGER: Final[logging.Logger] = logging.getLogger(__name__)

DEFAULT_QUOTE_CHARACTERS: Final[Tuple[str, str]] = ('"', "'")


@overload
def to_lower_if_not_quoted(value: str, quote_characters: Sequence[str] = ...) -> str: ...


@overload
def to_lower_if_not_quoted(value: None, quote_characters: Sequence[str] = ...) -> None: ...


def to_lower_if_not_quoted(
    value: str | None,
    quote_characters: Sequence[str] = DEFAULT_QUOTE_CHARACTERS,
) -> str | None:
    """
    Convert a string to lowercase if it is not enclosed in quotes.
    """
    if not value:
        return value
    for char in quote_characters:
        if value.startswith(char) and value.endswith(char):
            LOGGER.warning(
                f"The {value} string is bracketed by quotes,"
                " so it will not be converted to lowercase."
                " May cause sqlalchemy case-sensitivity issues."
            )
            return value
    LOGGER.info(f"Setting {value} to lowercase to ensure sqlalchemy case-insensitivity.")
    return value.lower()


class SQLDatasourceError(Exception):
    pass


class SQLAlchemyCreateEngineError(SQLDatasourceError):
    """
    An error creating a SQLAlchemy `Engine` object.

    Not to be confused with the GX `SQLAlchemyExecutionEngine`.
    """

    @overload
    def __init__(self, addendum: str | None = ..., cause: Exception = ...): ...

    @overload
    def __init__(self, addendum: str = ..., cause: Exception | None = ...): ...

    def __init__(
        self,
        addendum: str | None = None,
        cause: Exception | None = None,
    ):
        """Must provide a `cause`, `addendum`, or both."""
        message = "Unable to create SQLAlchemy Engine"
        if cause:
            message += f": due to {cause!r}"
        if addendum:
            message += f": {addendum}"
        super().__init__(message)


class _Partitioner(PartitionerProtocol, Protocol):
    def param_defaults(self, sql_asset: _SQLAsset) -> List[Dict]:
        """Creates all valid batch requests options for sql_asset

        This can be implemented by querying the data defined in the sql_asset to generate
        all the possible parameter values for the BatchRequest.options that will return data.
        For example for a YearMonth partitioner, we can query the underlying data to return the
        set of distinct (year, month) pairs. We would then return a list of BatchRequest.options,
        ie dictionaries, of the form {"year": year, "month": month} that contain all these distinct
        pairs.
        """
        ...


def _partitioner_and_sql_asset_to_batch_identifier_data(
    partitioner: _Partitioner, asset: _SQLAsset
) -> list[dict]:
    execution_engine = asset.datasource.get_execution_engine()
    sqlalchemy_data_partitioner = SqlAlchemyDataPartitioner(execution_engine.dialect_name)
    return sqlalchemy_data_partitioner.get_data_for_batch_identifiers(
        execution_engine=execution_engine,
        selectable=asset.as_selectable(),
        partitioner_method_name=partitioner.method_name,
        partitioner_kwargs=partitioner.partitioner_method_kwargs(),
    )


class _PartitionerDatetime(FluentBaseModel):
    column_name: str
    method_name: str
    sort_ascending: bool = True

    @property
    def columns(self) -> list[str]:
        return [self.column_name]

    def param_defaults(self, sql_asset: _SQLAsset) -> list[dict]:
        batch_identifier_data = _partitioner_and_sql_asset_to_batch_identifier_data(
            partitioner=self, asset=sql_asset
        )
        params: list[dict] = []
        for identifer_data in batch_identifier_data:
            params.append(identifer_data[self.column_name])
        return params

    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        """Validates all the datetime parameters for this partitioner exist in `options`."""
        identifiers: Dict = {}
        for part in self.param_names:
            if part not in options:
                raise ValueError(f"'{part}' must be specified in the batch parameters")  # noqa: TRY003 # FIXME CoP
            identifiers[part] = options[part]
        return {self.column_name: identifiers}

    @property
    def param_names(self) -> list[str]:
        raise NotImplementedError

    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class SqlPartitionerYear(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year"] = "partition_on_year"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}


class SqlPartitionerYearAndMonth(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month"] = "partition_on_year_and_month"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year", "month"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}


class SqlPartitionerYearAndMonthAndDay(_PartitionerDatetime):
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_year_and_month_and_day"] = (
        "partition_on_year_and_month_and_day"
    )

    @property
    @override
    def param_names(self) -> List[str]:
        return ["year", "month", "day"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}


class SqlPartitionerDatetimePart(_PartitionerDatetime):
    datetime_parts: List[str]
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_date_parts"] = "partition_on_date_parts"

    @property
    @override
    def param_names(self) -> List[str]:
        return self.datetime_parts

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name, "date_parts": self.param_names}

    @pydantic.validator("datetime_parts", each_item=True)
    def _check_param_name_allowed(cls, v: str):
        allowed_date_parts = [part.value for part in DatePart]
        assert v in allowed_date_parts, (
            f"Only the following param_names are allowed: {allowed_date_parts}"
        )
        return v


class _PartitionerOneColumnOneParam(FluentBaseModel):
    column_name: str
    method_name: str
    sort_ascending: bool = True

    @property
    def columns(self) -> list[str]:
        return [self.column_name]

    def param_defaults(self, sql_asset: _SQLAsset) -> list[dict]:
        batch_identifier_data = _partitioner_and_sql_asset_to_batch_identifier_data(
            partitioner=self, asset=sql_asset
        )
        params: list[dict] = []
        for identifer_data in batch_identifier_data:
            params.append({self.param_names[0]: identifer_data[self.column_name]})
        return params

    @property
    def param_names(self) -> list[str]:
        raise NotImplementedError

    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        raise NotImplementedError


class SqlPartitionerDividedInteger(_PartitionerOneColumnOneParam):
    divisor: int
    column_name: str
    method_name: Literal["partition_on_divided_integer"] = "partition_on_divided_integer"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["quotient"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name, "divisor": self.divisor}

    @override
    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if "quotient" not in options:
            raise ValueError("'quotient' must be specified in the batch parameters")  # noqa: TRY003 # FIXME CoP
        return {self.column_name: options["quotient"]}


class SqlPartitionerModInteger(_PartitionerOneColumnOneParam):
    mod: int
    column_name: str
    method_name: Literal["partition_on_mod_integer"] = "partition_on_mod_integer"

    @property
    @override
    def param_names(self) -> List[str]:
        return ["remainder"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name, "mod": self.mod}

    @override
    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if "remainder" not in options:
            raise ValueError("'remainder' must be specified in the batch parameters")  # noqa: TRY003 # FIXME CoP
        return {self.column_name: options["remainder"]}


class SqlPartitionerColumnValue(_PartitionerOneColumnOneParam):
    column_name: str
    method_name: Literal["partition_on_column_value"] = "partition_on_column_value"

    @property
    @override
    def param_names(self) -> List[str]:
        return [self.column_name]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}

    @override
    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if self.column_name not in options:
            raise ValueError(f"'{self.column_name}' must be specified in the batch parameters")  # noqa: TRY003 # FIXME CoP
        return {self.column_name: options[self.column_name]}

    @override
    def param_defaults(self, sql_asset: _SQLAsset) -> list[dict]:
        # The superclass version of param_defaults is correct, but here we leverage that
        # the parameter name is the same as the column name to make this much faster.
        return _partitioner_and_sql_asset_to_batch_identifier_data(
            partitioner=self, asset=sql_asset
        )


class SqlPartitionerMultiColumnValue(FluentBaseModel):
    column_names: List[str]
    sort_ascending: bool = True
    method_name: Literal["partition_on_multi_column_values"] = "partition_on_multi_column_values"

    @property
    def columns(self):
        return self.column_names

    @property
    def param_names(self) -> List[str]:
        return self.column_names

    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {"column_names": self.column_names}

    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if not (set(self.column_names) <= set(options.keys())):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"All column names, {self.column_names}, must be specified in the batch parameters. "  # noqa: E501 # FIXME CoP
                f" The options provided were f{options}."
            )
        return {col: options[col] for col in self.column_names}

    def param_defaults(self, sql_asset: _SQLAsset) -> list[dict]:
        return _partitioner_and_sql_asset_to_batch_identifier_data(
            partitioner=self, asset=sql_asset
        )


class SqlitePartitionerConvertedDateTime(_PartitionerOneColumnOneParam):
    """A partitioner than can be used for sql engines that represents datetimes as strings.

    The SQL engine that this currently supports is SQLite since it stores its datetimes as
    strings.
    The DatetimePartitioner will also work for SQLite and may be more intuitive.
    """

    # date_format_strings syntax is documented here:
    # https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    # It allows for arbitrary strings so can't be validated until conversion time.
    date_format_string: str
    column_name: str
    sort_ascending: bool = True
    method_name: Literal["partition_on_converted_datetime"] = "partition_on_converted_datetime"

    @property
    @override
    def param_names(self) -> List[str]:
        # The datetime parameter will be a string representing a datetime in the format
        # given by self.date_format_string.
        return ["datetime"]

    @override
    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        return {
            "column_name": self.column_name,
            "date_format_string": self.date_format_string,
        }

    @override
    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        if "datetime" not in options:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "'datetime' must be specified in the batch parameters to create a batch identifier"
            )
        return {self.column_name: options["datetime"]}


# We create this type instead of using _Partitioner so pydantic can use to this to
# coerce the partitioner to the right type during deserialization from config.
SqlPartitioner = Union[
    SqlPartitionerColumnValue,
    SqlPartitionerMultiColumnValue,
    SqlPartitionerDividedInteger,
    SqlPartitionerModInteger,
    SqlPartitionerYear,
    SqlPartitionerYearAndMonth,
    SqlPartitionerYearAndMonthAndDay,
    SqlPartitionerDatetimePart,
    SqlitePartitionerConvertedDateTime,
]


@public_api
class _SQLAsset(DataAsset[DatasourceT, ColumnPartitioner], Generic[DatasourceT]):
    """A _SQLAsset Mixin

    This is used as a mixin for _SQLAsset subclasses to give them the TableAsset functionality
    that can be used by different SQL datasource subclasses.

    For example see TableAsset defined in this module and SqliteTableAsset defined in
    sqlite_datasource.py
    """

    # Instance fields
    type: str = pydantic.Field("_sql_asset")
    name: str
    _partitioner_implementation_map: Dict[
        Type[ColumnPartitioner], Optional[Type[SqlPartitioner]]
    ] = pydantic.PrivateAttr(
        default={
            ColumnPartitionerYearly: SqlPartitionerYear,
            ColumnPartitionerMonthly: SqlPartitionerYearAndMonth,
            ColumnPartitionerDaily: SqlPartitionerYearAndMonthAndDay,
            PartitionerColumnValue: SqlPartitionerColumnValue,
            PartitionerDatetimePart: SqlPartitionerDatetimePart,
            PartitionerDividedInteger: SqlPartitionerDividedInteger,
            PartitionerModInteger: SqlPartitionerModInteger,
            PartitionerMultiColumnValue: SqlPartitionerMultiColumnValue,
            PartitionerConvertedDatetime: None,  # only implemented for sqlite backend
        }
    )

    def get_partitioner_implementation(
        self, abstract_partitioner: ColumnPartitioner
    ) -> SqlPartitioner:
        PartitionerClass = self._partitioner_implementation_map.get(type(abstract_partitioner))
        if not PartitionerClass:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Requested Partitioner `{abstract_partitioner.method_name}` is not implemented for this DataAsset. "  # noqa: E501 # FIXME CoP
            )
        assert PartitionerClass is not None
        return PartitionerClass(**abstract_partitioner.dict())

    @override
    def get_batch_parameters_keys(
        self,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> tuple[str, ...]:
        option_keys: Tuple[str, ...] = tuple()
        if partitioner:
            sql_partitioner = self.get_partitioner_implementation(partitioner)
            option_keys += tuple(sql_partitioner.param_names)
        return option_keys

    @override
    def test_connection(self) -> None:
        pass

    @staticmethod
    def _matches_request_options(candidate: Dict, requested_options: BatchParameters) -> bool:
        for k, v in requested_options.items():
            if isinstance(candidate[k], (datetime, date)):
                candidate[k] = str(candidate[k])

            if v is not None and candidate[k] != v:
                return False
        return True

    def _fully_specified_batch_requests(self, batch_request: BatchRequest) -> List[BatchRequest]:
        """Populates a batch requests unspecified params producing a list of batch requests."""

        if batch_request.partitioner is None:
            # Currently batch_request.options is complete determined by the presence of a
            # partitioner. If partitioner is None, then there are no specifiable options
            # so we return early. Since the passed in batch_request is verified, it must be the
            # empty, ie {}.
            # In the future, if there are options that are not determined by the partitioner
            # this check will have to be generalized.
            return [batch_request]

        sql_partitioner = self.get_partitioner_implementation(batch_request.partitioner)

        batch_requests: List[BatchRequest] = []
        # We iterate through all possible batches as determined by the partitioner
        for params in sql_partitioner.param_defaults(self):
            # If the params from the partitioner don't match the batch parameters
            # we don't create this batch.
            if not _SQLAsset._matches_request_options(params, batch_request.options):
                continue
            options = copy.deepcopy(batch_request.options)
            options.update(params)
            batch_requests.append(
                BatchRequest(
                    datasource_name=batch_request.datasource_name,
                    data_asset_name=batch_request.data_asset_name,
                    options=options,
                    partitioner=batch_request.partitioner,
                )
            )
        return batch_requests

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        self._validate_batch_request(batch_request)
        if batch_request.partitioner:
            sql_partitioner = self.get_partitioner_implementation(batch_request.partitioner)
        else:
            sql_partitioner = None

        requests = self._fully_specified_batch_requests(batch_request)
        metadata_dicts = [self._get_batch_metadata_from_batch_request(r) for r in requests]

        if sql_partitioner:
            metadata_dicts = self.sort_batch_identifiers_list(metadata_dicts, sql_partitioner)

        return metadata_dicts[batch_request.batch_slice]

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        """Batch that matches the BatchRequest.

        Args:
            batch_request: A batch request for this asset. Usually obtained by calling
                build_batch_request on the asset.

        Returns:
            A list Batch that matches the options specified in the batch request.
        """
        self._validate_batch_request(batch_request)

        if batch_request.partitioner:
            sql_partitioner = self.get_partitioner_implementation(batch_request.partitioner)
        else:
            sql_partitioner = None

        batch_spec_kwargs: dict[str, str | dict | None]
        requests = self._fully_specified_batch_requests(batch_request)
        unsorted_metadata_dicts = [self._get_batch_metadata_from_batch_request(r) for r in requests]

        if not unsorted_metadata_dicts:
            raise NoAvailableBatchesError()

        if sql_partitioner:
            sorted_metadata_dicts = self.sort_batch_identifiers_list(
                unsorted_metadata_dicts, sql_partitioner
            )
        else:
            sorted_metadata_dicts = unsorted_metadata_dicts

        sorted_metadata_dicts = sorted_metadata_dicts[batch_request.batch_slice]
        batch_metadata = sorted_metadata_dicts[-1]

        # we've sorted the metadata, but not the requests, so we need the index of our
        # batch_metadata from the original unsorted list so that we get the right request
        request_index = unsorted_metadata_dicts.index(batch_metadata)

        request = requests[request_index]
        batch_spec_kwargs = self._create_batch_spec_kwargs()
        if sql_partitioner:
            batch_spec_kwargs["partitioner_method"] = sql_partitioner.method_name
            batch_spec_kwargs["partitioner_kwargs"] = sql_partitioner.partitioner_method_kwargs()
            # mypy infers that batch_spec_kwargs["batch_identifiers"] is a collection, but
            # it is hardcoded to a dict above, so we cast it here.
            cast("Dict", batch_spec_kwargs["batch_identifiers"]).update(
                sql_partitioner.batch_parameters_to_batch_spec_kwarg_identifiers(request.options)
            )
        # Creating the batch_spec is our hook into the execution engine.
        batch_spec = self._create_batch_spec(batch_spec_kwargs)
        execution_engine: SqlAlchemyExecutionEngine = self.datasource.get_execution_engine()
        data, markers = execution_engine.get_batch_data_and_markers(batch_spec=batch_spec)

        batch_definition = LegacyBatchDefinition(
            datasource_name=self.datasource.name,
            data_connector_name=_DATA_CONNECTOR_NAME,
            data_asset_name=self.name,
            batch_identifiers=IDDict(batch_spec["batch_identifiers"]),
            batch_spec_passthrough=None,
        )

        return Batch(
            datasource=self.datasource,
            data_asset=self,
            batch_request=request,
            data=data,
            metadata=batch_metadata,
            batch_markers=markers,
            batch_spec=batch_spec,
            batch_definition=batch_definition,
        )

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> BatchRequest:
        """A batch request that can be used to obtain batches for this DataAsset.

        Args:
            options: A dict that can be used to filter the batch groups returned from the asset.
                The dict structure depends on the asset type. The available keys for dict can be obtained by
                calling batch_parameters.
            batch_slice: A python slice that can be used to limit the sorted batches by index.
                e.g. `batch_slice = "[-5:]"` will request only the last 5 batches after the options filter is applied.
            partitioner: A Partitioner used to narrow the data returned from the asset.

        Returns:
            A BatchRequest object that can be used to obtain a batch from an Asset by calling the
            get_batch method.
        """  # noqa: E501 # FIXME CoP
        if options is not None and not self._batch_parameters_are_valid(
            options=options, partitioner=partitioner
        ):
            allowed_keys = set(self.get_batch_parameters_keys(partitioner=partitioner))
            actual_keys = set(options.keys())
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "batch parameters should only contain keys from the following set:\n"
                f"{allowed_keys}\nbut your specified keys contain\n"
                f"{actual_keys.difference(allowed_keys)}\nwhich is not valid.\n"
            )

        return BatchRequest(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            options=options or {},
            batch_slice=batch_slice,
            partitioner=partitioner,
        )

    @override
    def add_batch_definition(
        self,
        name: str,
        partitioner: Optional[ColumnPartitioner] = None,
        validate_partitioner: bool = True,
    ) -> BatchDefinition[ColumnPartitioner]:
        if validate_partitioner and partitioner:
            self.validate_batch_definition(partitioner)
        return super().add_batch_definition(name, partitioner)

    @public_api
    def validate_batch_definition(self, partitioner: ColumnPartitioner) -> None:
        """Validates that the Batch Definition column is of a permissible type

        This isn't meant to be called directly. This is called internally when a Batch Definition
         is added. Data asset implementers can override this for their specific data asset.

        Raises:
            SqlAddBatchDefinitionError: The specified column to partition on is not of
                a permissible type for batching (ie date or datetime) or no data is
                present in this column.
        """
        # We only support certain partitioners for using as batch definitions.
        assert isinstance(
            partitioner,
            (
                ColumnPartitionerYearly,
                ColumnPartitionerMonthly,
                ColumnPartitionerDaily,
            ),
        )
        # A _SQLAsset must have a SQLDatasource
        assert isinstance(self.datasource, SQLDatasource)
        engine: sqlalchemy.Engine = self.datasource.get_engine()

        # It would be better to introspect the database types and see which ones map to date or
        # datetime. However, 3rd party types, such as Snowflakes TIMESTAMP_NTZ haven't implemented
        # all the sqlalchemy abstract methods such as `python_type`.
        # To make this more concrete I would have liked to do something like:
        # insp = sqlalchemy.inspect(self.datasource.get_engine())
        # cols = insp.get_columns(self.table_name, self.schema_name)
        # for col in cols:
        #     pytype = col['type'].python_type
        #
        # Instead we query the db for a non-null value to see if sqlalchemy converts
        # this value to a python date or datetime. This means we REQUIRE that data is
        # present for this validation to work.
        with engine.connect() as connection:
            selectable: sqlalchemy.Selectable = self.as_selectable()
            column: sqlalchemy.ColumnClause[Never] = sa.sql.column(partitioner.column_name)
            try:
                row = connection.execute(
                    sa.select(column, selectable).limit(1)  # type: ignore[call-overload]  # sqlalchemy typing is missing variants
                )
            except Exception as query_error:
                raise SqlAddBatchDefinitionError(
                    msg=f"Attempt to read an example non-null '{column}' value from '{selectable}'"
                    " failed so column type can't be verified to be a date or datetime."
                ) from query_error

            r = row.first()
            if not r or not isinstance(getattr(r, partitioner.column_name), (datetime, date)):
                raise SqlAddBatchDefinitionError(
                    msg=f"'{column}' column from '{selectable}' is not a date or datetime type."
                )

    @public_api
    def add_batch_definition_whole_table(self, name: str) -> BatchDefinition:
        """Adds a whole table Batch Definition to this Data Asset

        Args:
            name: The name of the Batch Definition to be added

        Returns:
            The added BatchDefinition object.
        """
        return self.add_batch_definition(
            name=name,
            partitioner=None,
        )

    @public_api
    def add_batch_definition_yearly(
        self,
        name: str,
        column: str,
        sort_ascending: bool = True,
        validate_batchable: bool = True,
    ) -> BatchDefinition:
        """Adds a yearly Batch Definition to this Data Asset

        Args:
            name: The name of the Batch Definition to be added.
            column: The column name on which to partition the asset by year.
            sort_ascending: Boolean to indicate whether to sort ascending (default) or descending.
                When running a validation, we default to running the last Batch Definition
                if one is not explicitly specified.

        Returns:
            The added BatchDefinition object.
        """

        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerYearly(
                method_name="partition_on_year", column_name=column, sort_ascending=sort_ascending
            ),
            validate_partitioner=validate_batchable,
        )

    @public_api
    def add_batch_definition_monthly(
        self,
        name: str,
        column: str,
        sort_ascending: bool = True,
        validate_batchable: bool = True,
    ) -> BatchDefinition:
        """Adds a monthly Batch Definition to this Data Asset

        Args:
            name: The name of the Batch Definition to be added
            column: The column name on which to partition the asset by month
            sort_ascending: Boolean to indicate whether to sort ascending (default) or descending.
                When running a validation, we default to running the last Batch Definition
                if one is not explicitly specified.

        Returns:
            The added BatchDefinition object.
        """

        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerMonthly(
                method_name="partition_on_year_and_month",
                column_name=column,
                sort_ascending=sort_ascending,
            ),
            validate_partitioner=validate_batchable,
        )

    @public_api
    def add_batch_definition_daily(
        self,
        name: str,
        column: str,
        sort_ascending: bool = True,
        validate_batchable: bool = True,
    ) -> BatchDefinition:
        """Adds a daily Batch Definition to this Data Asset

        Args:
            name: The name of the Batch Definition to be added
            column: The column name on which to partition the asset by day
            sort_ascending: Boolean to indicate whether to sort ascending (default) or descending.
                When running a validation, we default to running the last Batch Definition
                if one is not explicitly specified.

        Returns:
            The added BatchDefinition object.
        """

        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerDaily(
                method_name="partition_on_year_and_month_and_day",
                column_name=column,
                sort_ascending=sort_ascending,
            ),
            validate_partitioner=validate_batchable,
        )

    @override
    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        if not (
            batch_request.datasource_name == self.datasource.name
            and batch_request.data_asset_name == self.name
            and self._batch_parameters_are_valid(
                options=batch_request.options,
                partitioner=batch_request.partitioner,
            )
        ):
            options = dict.fromkeys(
                self.get_batch_parameters_keys(partitioner=batch_request.partitioner)
            )
            expect_batch_request_form = BatchRequest[ColumnPartitioner](
                datasource_name=self.datasource.name,
                data_asset_name=self.name,
                options=options,
                batch_slice=batch_request._batch_slice_input,  # type: ignore[attr-defined] # FIXME CoP
                partitioner=batch_request.partitioner,
            )
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "BatchRequest should have form:\n"
                f"{pf(expect_batch_request_form.dict())}\n"
                f"but actually has form:\n{pf(batch_request.dict())}\n"
            )

    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        """Creates batch_spec_kwargs used to instantiate a SqlAlchemyDatasourceBatchSpec or RuntimeQueryBatchSpec

        This is called by get_batch to generate the batch.

        Returns:
            A dictionary that will be passed to self._create_batch_spec(**returned_dict)
        """  # noqa: E501 # FIXME CoP
        raise NotImplementedError

    def _create_batch_spec(self, batch_spec_kwargs: dict) -> BatchSpec:
        """
        Instantiates a SqlAlchemyDatasourceBatchSpec or RuntimeQueryBatchSpec.
        """
        raise NotImplementedError

    def as_selectable(self) -> sqlalchemy.Selectable:
        """Returns a Selectable that can be used to query this data

        Returns:
            A Selectable that can be used in a from clause to query this data
        """
        raise NotImplementedError


@public_api
class QueryAsset(_SQLAsset):
    """An asset made from a SQL query

    Args:
        query: The query to be used to construct the underlying Data Asset
    """

    # Instance fields
    type: Literal["query"] = "query"
    query: str

    @pydantic.validator("query")
    def query_must_start_with_select(cls, v: str):
        query = v.lstrip()
        if not (query.upper().startswith("SELECT") and query[6].isspace()):
            raise ValueError("query must start with 'SELECT' followed by a whitespace.")  # noqa: TRY003 # FIXME CoP
        return v

    @override
    def as_selectable(self) -> sqlalchemy.Selectable:
        """Returns the Selectable that is used to retrieve the data.

        This can be used in a subselect FROM clause for queries against this data.
        """
        return sa.select(sa.text(self.query.lstrip()[6:])).subquery()

    @override
    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        return {
            "data_asset_name": self.name,
            "query": self.query,
            "temp_table_schema_name": None,
            "batch_identifiers": {},
        }

    @override
    def _create_batch_spec(self, batch_spec_kwargs: dict) -> RuntimeQueryBatchSpec:
        return RuntimeQueryBatchSpec(**batch_spec_kwargs)


@public_api
class TableAsset(_SQLAsset):
    """A class representing a table from a SQL database

    Args:
        table_name: The name of the database table to be added
        schema_name: The name of the schema containing the database table to be added.
    """

    # Instance fields
    type: Literal["table"] = "table"
    # TODO: quoted_name or str
    table_name: str = pydantic.Field(
        "",
        description="Name of the SQL table. Will default to the value of `name` if not provided.",
    )
    schema_name: Optional[str] = None

    @property
    def qualified_name(self) -> str:
        return f"{self.schema_name}.{self.table_name}" if self.schema_name else self.table_name

    @pydantic.validator("table_name", pre=True, always=True)
    def _default_table_name(cls, table_name: str, values: dict, **kwargs) -> str:
        if not (validated_table_name := table_name or values.get("name")):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "table_name cannot be empty and should default to name if not provided"
            )

        return validated_table_name

    @pydantic.validator("table_name")
    def _resolve_quoted_name(cls, table_name: str) -> str | quoted_name:
        table_name_is_quoted: bool = cls._is_bracketed_by_quotes(table_name)

        # We reimport sqlalchemy from our compatability layer because we make
        # quoted_name a top level import there.
        from great_expectations.compatibility import sqlalchemy

        if sqlalchemy.quoted_name:  # type: ignore[truthy-function] # FIXME CoP
            if isinstance(table_name, sqlalchemy.quoted_name):
                return table_name

            if table_name_is_quoted:
                # https://docs.sqlalchemy.org/en/20/core/sqlelement.html#sqlalchemy.sql.expression.quoted_name.quote
                # Remove the quotes and add them back using the sqlalchemy.quoted_name function
                # TODO: We need to handle nested quotes
                table_name = table_name.strip("'").strip('"')

            return sqlalchemy.quoted_name(
                value=table_name,
                quote=table_name_is_quoted,
            )

        return table_name

    @override
    def test_connection(self) -> None:
        """Test the connection for the TableAsset.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        datasource: SQLDatasource = self.datasource
        engine: sqlalchemy.Engine = datasource.get_engine()
        inspector: sqlalchemy.Inspector = sa.inspect(engine)

        if self.schema_name and self.schema_name not in inspector.get_schema_names():
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f'Attempt to connect to table: "{self.qualified_name}" failed because the schema '
                f'"{self.schema_name}" does not exist.'
            )

        try:
            with engine.connect() as connection:
                table = sa.table(self.table_name, schema=self.schema_name)
                # don't need to fetch any data, just want to make sure the table is accessible
                connection.execute(sa.select(1, table).limit(1))
        except Exception as query_error:
            LOGGER.info(f"{self.name} `.test_connection()` query failed: {query_error!r}")
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f"Attempt to connect to table: {self.qualified_name} failed because the test query "
                f"failed. Ensure the table exists and the user has access to select data from the table: {query_error}"  # noqa: E501 # FIXME CoP
            ) from query_error

    @override
    def as_selectable(self) -> sqlalchemy.Selectable:
        """Returns the table as a sqlalchemy Selectable.

        This can be used in a from clause for a query against this data.
        """
        return sa.table(self.table_name, schema=self.schema_name)

    @override
    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        return {
            "type": "table",
            "data_asset_name": self.name,
            "table_name": self.table_name,
            "schema_name": self.schema_name,
            "batch_identifiers": {},
        }

    @override
    def _create_batch_spec(self, batch_spec_kwargs: dict) -> SqlAlchemyDatasourceBatchSpec:
        return SqlAlchemyDatasourceBatchSpec(**batch_spec_kwargs)

    @staticmethod
    def _is_bracketed_by_quotes(target: str) -> bool:
        """
        Returns True if the target string is bracketed by quotes.

        Override this method if the quote characters are different than `'` or `"` in the
        target database, such as backticks in Databricks SQL.

        Arguments:
            target: A string to check if it is bracketed by quotes.

        Returns:
            True if the target string is bracketed by quotes.
        """
        return any(
            target.startswith(quote) and target.endswith(quote)
            for quote in DEFAULT_QUOTE_CHARACTERS
        )

    @classmethod
    def _to_lower_if_not_bracketed_by_quotes(cls, target: str) -> str:
        """Returns the target string in lowercase if it is not bracketed by quotes.
        This is used to ensure case-insensitivity in sqlalchemy queries.

        Arguments:
            target: A string to convert to lowercase if it is not bracketed by quotes.

        Returns:
            The target string in lowercase if it is not bracketed by quotes.
        """
        return to_lower_if_not_quoted(target, quote_characters=DEFAULT_QUOTE_CHARACTERS)


def _warn_for_more_specific_datasource_type(connection_string: str) -> None:
    """
    Warns if a more specific datasource type may be more appropriate based on the connection string connector prefix.
    """  # noqa: E501 # FIXME CoP
    from great_expectations.datasource.fluent.sources import DataSourceManager

    connector: str = connection_string.split("://")[0].split("+")[0]

    type_lookup_plus: dict[str, str] = {
        n: DataSourceManager.type_lookup[n].__name__
        for n in DataSourceManager.type_lookup.type_names()
    }
    # type names are not always exact match to connector strings
    type_lookup_plus.update(
        {
            "postgresql": type_lookup_plus["postgres"],
            "databricks": type_lookup_plus["databricks_sql"],
        }
    )

    more_specific_datasource: str | None = type_lookup_plus.get(connector)
    if more_specific_datasource:
        warnings.warn(
            f"You are using a generic SQLDatasource but a more specific {more_specific_datasource} "
            "may be more appropriate"
            " https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/fluent/database/connect_sql_source_data",
            category=GxDatasourceWarning,
        )


# This improves our error messages by providing a more specific type for pydantic to validate against  # noqa: E501 # FIXME CoP
# It also ensure the generated jsonschema has a oneOf instead of anyOf field for assets
# https://docs.pydantic.dev/1.10/usage/types/#discriminated-unions-aka-tagged-unions
AssetTypes = Annotated[Union[TableAsset, QueryAsset], Field(discriminator="type")]


@public_api
class SQLDatasource(Datasource):
    """Adds a generic SQL datasource to the data context.

    Args:
        name: The name of this datasource.
        connection_string: The SQLAlchemy connection string used to connect to the database.
            For example: "postgresql+psycopg2://postgres:@localhost/test_database"
        create_temp_table: Whether to leverage temporary tables during metric computation.
        kwargs: Extra SQLAlchemy keyword arguments to pass to `create_engine()`. Note, only python
            primitive types will be serializable to config.
        assets: An optional dictionary whose keys are SQL DataAsset names and whose values
            are SQL DataAsset objects.
    """

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [TableAsset, QueryAsset]

    # right side of the operator determines the type name
    # left side enforces the names on instance creation
    type: Literal["sql"] = "sql"
    connection_string: Union[ConfigStr, str]
    create_temp_table: bool = False
    kwargs: Dict[str, Union[ConfigStr, Any]] = pydantic.Field(
        default={},
        description="Optional dictionary of `kwargs` will be passed to the SQLAlchemy Engine"
        " as part of `create_engine(connection_string, **kwargs)`",
    )
    # We need to explicitly add each asset type to the Union due to how
    # deserialization is implemented in our pydantic base model.
    assets: List[AssetTypes] = []

    # private attrs
    _cached_connection_string: Union[str, ConfigStr] = pydantic.PrivateAttr("")
    _engine: Union[sqlalchemy.Engine, None] = pydantic.PrivateAttr(None)

    # These are instance var because ClassVars can't contain Type variables. See
    # https://peps.python.org/pep-0526/#class-and-instance-variable-annotations
    _TableAsset: Type[TableAsset] = pydantic.PrivateAttr(TableAsset)
    _QueryAsset: Type[QueryAsset] = pydantic.PrivateAttr(QueryAsset)

    @property
    @override
    def execution_engine_type(self) -> Type[SqlAlchemyExecutionEngine]:
        """Returns the default execution engine type."""
        return SqlAlchemyExecutionEngine

    def get_engine(self) -> sqlalchemy.Engine:
        if self.connection_string != self._cached_connection_string or not self._engine:
            try:
                self._engine = self._create_engine()
            except Exception as e:
                # connection_string has passed pydantic validation, but still fails to create a sqlalchemy engine  # noqa: E501 # FIXME CoP
                # one possible case is a missing plugin (e.g. psycopg2)
                raise SQLAlchemyCreateEngineError(cause=e) from e
            self._cached_connection_string = self.connection_string
        return self._engine

    def _create_engine(self) -> sqlalchemy.Engine:
        model_dict = self.dict(
            exclude=self._get_exec_engine_excludes(),
            config_provider=self._config_provider,
        )
        _check_config_substitutions_needed(
            self, model_dict, raise_warning_if_provider_not_present=True
        )
        # the connection_string has had config substitutions applied
        connection_string = model_dict.pop("connection_string")
        if self.__class__.__name__ == "SQLDatasource":
            _warn_for_more_specific_datasource_type(connection_string)
        kwargs = model_dict.pop("kwargs", {})
        return sa.create_engine(connection_string, **kwargs)

    @override
    def get_execution_engine(self) -> SqlAlchemyExecutionEngine:
        # Overrides get_execution_engine in Datasource
        # because we need to pass the kwargs as keyvalue args to the execution engine
        # when then passes them to the engine.
        current_execution_engine_kwargs = self.dict(
            exclude=self._get_exec_engine_excludes(),
            config_provider=self._config_provider,
            # by default we exclude unset values to prevent lots of extra values in the yaml files
            # but we want to include them here
            exclude_unset=False,
        )
        if (
            current_execution_engine_kwargs != self._cached_execution_engine_kwargs
            or not self._execution_engine
        ):
            self._cached_execution_engine_kwargs = current_execution_engine_kwargs
            engine_kwargs = current_execution_engine_kwargs.pop("kwargs", {})
            self._execution_engine = self._execution_engine_type()(
                **current_execution_engine_kwargs,
                **engine_kwargs,
            )
        return self._execution_engine

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the SQLDatasource.

        Args:
            test_assets: If assets have been passed to the SQLDatasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """  # noqa: E501 # FIXME CoP
        try:
            engine: sqlalchemy.Engine = self.get_engine()
            engine.connect()
        except Exception as e:
            raise TestConnectionError(cause=e) from e
        if self.assets and test_assets:
            for asset in self.assets:
                asset._datasource = self
                asset.test_connection()

    @public_api
    def add_table_asset(
        self,
        name: str,
        table_name: str = "",
        schema_name: Optional[str] = None,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> TableAsset:
        """Adds a table asset to this datasource.

        Args:
            name: The name of this table asset.
            table_name: The table where the data resides.
            schema_name: The schema that holds the table.
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches derived from it.

        Returns:
            The table asset that is added to the datasource.
            The type of this object will match the necessary type for this datasource.
            eg, it could be a TableAsset or a SqliteTableAsset.
        """  # noqa: E501 # FIXME CoP
        if schema_name:
            schema_name = self._TableAsset._to_lower_if_not_bracketed_by_quotes(schema_name)
        asset = self._TableAsset(
            name=name,
            table_name=table_name,
            schema_name=schema_name,
            batch_metadata=batch_metadata or {},
        )
        return self._add_asset(asset)

    @public_api
    def add_query_asset(
        self,
        name: str,
        query: str,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> QueryAsset:
        """Adds a query asset to this datasource.

        Args:
            name: The name of this table asset.
            query: The SELECT query to selects the data to validate. It must begin with the "SELECT".
            batch_metadata: BatchMetadata we want to associate with this DataAsset and all batches derived from it.

        Returns:
            The query asset that is added to the datasource.
            The type of this object will match the necessary type for this datasource.
            eg, it could be a QueryAsset or a SqliteQueryAsset.
        """  # noqa: E501 # FIXME CoP
        asset = self._QueryAsset(
            name=name,
            query=query,
            batch_metadata=batch_metadata or {},
        )
        return self._add_asset(asset)



import pytest

from great_expectations.experimental.metric_repository.metrics import Metric

pytestmark = pytest.mark.unit


def test_cannot_init_abstract_metric():
    with pytest.raises(NotImplementedError):
        Metric(
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        )



import pandas as pd

from great_expectations.metrics.batch.sample_values import SampleValues, SampleValuesResult
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import ALL_DATA_SOURCES

COL_NAME = "my_column"

DATA_FRAME_WITH_MANY_ROWS = pd.DataFrame({COL_NAME: [i for i in range(100)]})
DATA_FRAME_WITH_FEW_ROWS = pd.DataFrame({COL_NAME: [1, 2, 3]})


class TestSampleValues:
    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME_WITH_MANY_ROWS,
    )
    def test_correct_return_types(self, batch_for_datasource) -> None:
        metric_result = batch_for_datasource.compute_metrics(SampleValues())
        assert isinstance(metric_result, SampleValuesResult)

        values = set(metric_result.value[COL_NAME])
        assert values.issubset(range(100))

    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME_WITH_MANY_ROWS,
    )
    def test_default_n_rows(self, batch_for_datasource) -> None:
        metric_result = batch_for_datasource.compute_metrics(SampleValues())
        assert isinstance(metric_result, SampleValuesResult)
        assert len(metric_result.value) == 10

    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME_WITH_FEW_ROWS,
    )
    def test_fewer_rows_exist_than_default_n_rows(self, batch_for_datasource) -> None:
        metric_result = batch_for_datasource.compute_metrics(SampleValues())
        assert isinstance(metric_result, SampleValuesResult)
        assert len(metric_result.value) == 3

    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME_WITH_MANY_ROWS,
    )
    def test_with_custom_n_rows(self, batch_for_datasource) -> None:
        n_rows = 5
        metric_result = batch_for_datasource.compute_metrics(SampleValues(n_rows=n_rows))
        assert isinstance(metric_result, SampleValuesResult)
        assert len(metric_result.value) == n_rows



"""Tests to ensure core validation flows work with sql assets

NOTE: assertions here take the form of asserting that expectations pass
based on knowledge of the data in the test set.
"""

import json
import pathlib
from typing import Dict, Optional

import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.datasource.fluent.sql_datasource import _SQLAsset
from great_expectations.self_check.util import (
    drop_table,
    get_test_validator_with_data,
)
from great_expectations.util import build_in_memory_runtime_context

pytestmark = pytest.mark.postgresql


DATABASE_NAME = "test_ci"
CONNECTION_STRING = f"postgresql+psycopg2://postgres:@localhost:5432/{DATABASE_NAME}"
TABLE_NAME = "ct_column_values_to_be_between__evaluation_parameters_dataset_1"


# constants for what we know about the test data
COLUMN_NAME = "x"
ALL_VALUES = list(range(1, 11))
INVALID_VALUES = list(range(1, 12))
VALUES_WITH_NO_DATE = [9]
VALUES_FOR_MOST_RECENT_DATE = [10]
MY_FAVORITE_DAY = {"year": 2000, "month": 6, "day": 1}
VALUES_ON_MY_FAVORITE_DAY = [8]


@pytest.fixture(scope="module", autouse=True)
def setup_module():
    # setup table
    dir_path = pathlib.Path(__file__).parent
    file = dir_path / "expect_column_values_to_be_between__evaluation_parameters.json"
    with open(file) as f:
        test_configuration = json.load(f)
    test_config = test_configuration["datasets"][0]
    pk_column = True
    schemas = test_config["schemas"]
    dataset = test_config["data"]
    # We call this for the side-effect of creating TABLE_NAME and loading dataset into it
    _ = get_test_validator_with_data(
        execution_engine="postgresql",
        data=dataset,
        table_name=TABLE_NAME,
        schemas=schemas,
        context=build_in_memory_runtime_context(),
        pk_column=pk_column,
    )
    yield
    drop_table(TABLE_NAME, CONNECTION_STRING)


@pytest.fixture
def context() -> AbstractDataContext:
    return gx.get_context(mode="ephemeral")


@pytest.fixture
def postgres_asset(context: AbstractDataContext) -> _SQLAsset:
    DATASOURCE_NAME = "postgres"
    ASSET_NAME = "ten trips"
    datasource = context.data_sources.add_postgres(
        DATASOURCE_NAME, connection_string=CONNECTION_STRING
    )
    data_asset = datasource.add_table_asset(name=ASSET_NAME, table_name=TABLE_NAME)

    return data_asset


@pytest.fixture
def postgres_whole_table_batch_definition(postgres_asset: _SQLAsset) -> BatchDefinition:
    return postgres_asset.add_batch_definition_whole_table("the whole table")


@pytest.fixture
def postgres_daily_batch_definition(postgres_asset: _SQLAsset) -> BatchDefinition:
    return postgres_asset.add_batch_definition_daily(name="daily", column="ts")


@pytest.fixture
def postgres_daily_batch_definition_descending(postgres_asset: _SQLAsset) -> BatchDefinition:
    return postgres_asset.add_batch_definition_daily(
        name="daily", column="ts", sort_ascending=False
    )


def _create_test_cases():
    """Create our test cases.

    With each flow, we want to see that we can validate an entire asset,
    as well as subsets of the asset, including sorting and using batch parameters.

    The positional arguments are:
    - An Expectation
    - The fixture name for the batch definition
    - The batch_parameters to pass in during validation
    """
    return [
        pytest.param(
            gxe.ExpectColumnDistinctValuesToEqualSet(column=COLUMN_NAME, value_set=ALL_VALUES),
            "postgres_whole_table_batch_definition",
            None,  # no batch parameters
            id="whole asset",
        ),
        pytest.param(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=COLUMN_NAME, value_set=VALUES_FOR_MOST_RECENT_DATE
            ),
            "postgres_daily_batch_definition",
            None,  # no batch parameters
            id="ascending",
        ),
        pytest.param(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=COLUMN_NAME, value_set=VALUES_WITH_NO_DATE
            ),
            "postgres_daily_batch_definition_descending",
            None,  # no batch parameters
            id="descending",
        ),
        pytest.param(
            gxe.ExpectColumnDistinctValuesToEqualSet(
                column=COLUMN_NAME, value_set=VALUES_ON_MY_FAVORITE_DAY
            ),
            "postgres_daily_batch_definition",
            MY_FAVORITE_DAY,
            id="batch params",
        ),
    ]


def test_get_batch_identifiers_list__simple(
    postgres_daily_batch_definition: BatchDefinition,
) -> None:
    batch_definition = postgres_daily_batch_definition
    batch_identifiers_list = batch_definition.get_batch_identifiers_list()

    assert len(batch_identifiers_list) == 10
    # just spot check the edges
    assert batch_identifiers_list[0] == {"year": None, "month": None, "day": None}
    assert batch_identifiers_list[-1] == {"year": 2001, "month": 1, "day": 1}


def test_get_batch_identifiers_list__respects_order(
    postgres_daily_batch_definition_descending: BatchDefinition,
) -> None:
    batch_definition = postgres_daily_batch_definition_descending
    batch_identifiers_list = batch_definition.get_batch_identifiers_list()

    assert len(batch_identifiers_list) == 10
    # just spot check the edges
    assert batch_identifiers_list[0] == {"year": 2001, "month": 1, "day": 1}
    assert batch_identifiers_list[-1] == {"year": None, "month": None, "day": None}


def test_get_batch_identifiers_list__respects_batch_params(
    postgres_daily_batch_definition: BatchDefinition,
) -> None:
    batch_definition = postgres_daily_batch_definition
    batch_identifiers_list = batch_definition.get_batch_identifiers_list(
        batch_parameters={"year": 2000}
    )

    assert len(batch_identifiers_list) == 6
    # just spot check the edges
    assert batch_identifiers_list[0] == {"year": 2000, "month": 1, "day": 1}
    assert batch_identifiers_list[-1] == {"year": 2000, "month": 6, "day": 1}


def test_get_batch_identifiers_list__no_batches(
    postgres_daily_batch_definition: BatchDefinition,
) -> None:
    batch_definition = postgres_daily_batch_definition
    batch_identifiers_list = batch_definition.get_batch_identifiers_list(
        batch_parameters={"year": 1900}
    )

    assert batch_identifiers_list == []


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_batch_validate_expectation(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    request: pytest.FixtureRequest,
) -> None:
    """Ensure Batch::validate(Epectation) works"""

    batch_definition: BatchDefinition = request.getfixturevalue(batch_definition_fixture_name)
    batch = batch_definition.get_batch(batch_parameters=batch_parameters)
    result = batch.validate(expectation)

    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_batch_validate_expectation_suite(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    request: pytest.FixtureRequest,
) -> None:
    """Ensure Batch::validate(EpectationSuite) works"""

    suite = ExpectationSuite("my suite", expectations=[expectation])
    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    batch = batch_definition.get_batch(batch_parameters=batch_parameters)
    result = batch.validate(suite)

    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_validation_definition_run(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    context: AbstractDataContext,
    request: pytest.FixtureRequest,
) -> None:
    """Ensure ValidationDefinition::run works"""

    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    suite = context.suites.add(ExpectationSuite("my_suite", expectations=[expectation]))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="whatever", data=batch_definition, suite=suite)
    )
    result = validation_definition.run(batch_parameters=batch_parameters)

    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_checkpoint_run(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    context: AbstractDataContext,
    request: pytest.FixtureRequest,
) -> None:
    """Ensure Checkpoint::run works"""

    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    suite = context.suites.add(ExpectationSuite("my_suite", expectations=[expectation]))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="whatever", data=batch_definition, suite=suite)
    )
    checkpoint = context.checkpoints.add(
        Checkpoint(name="whatever", validation_definitions=[validation_definition])
    )
    result = checkpoint.run(batch_parameters=batch_parameters)

    assert result.success



from great_expectations.expectations.expectation import Expectation

from .core import (
    ExpectColumnDistinctValuesToBeInSet,
    ExpectColumnDistinctValuesToContainSet,
    ExpectColumnDistinctValuesToEqualSet,
    ExpectColumnKLDivergenceToBeLessThan,
    ExpectColumnMaxToBeBetween,
    ExpectColumnMeanToBeBetween,
    ExpectColumnMedianToBeBetween,
    ExpectColumnMinToBeBetween,
    ExpectColumnMostCommonValueToBeInSet,
    ExpectColumnPairValuesAToBeGreaterThanB,
    ExpectColumnPairValuesToBeEqual,
    ExpectColumnPairValuesToBeInSet,
    ExpectColumnProportionOfNonNullValuesToBeBetween,
    ExpectColumnProportionOfUniqueValuesToBeBetween,
    ExpectColumnQuantileValuesToBeBetween,
    ExpectColumnStdevToBeBetween,
    ExpectColumnSumToBeBetween,
    ExpectColumnToExist,
    ExpectColumnUniqueValueCountToBeBetween,
    ExpectColumnValueLengthsToBeBetween,
    ExpectColumnValueLengthsToEqual,
    ExpectColumnValuesToBeBetween,
    ExpectColumnValuesToBeDateutilParseable,
    ExpectColumnValuesToBeDecreasing,
    ExpectColumnValuesToBeIncreasing,
    ExpectColumnValuesToBeInSet,
    ExpectColumnValuesToBeInTypeList,
    ExpectColumnValuesToBeJsonParseable,
    ExpectColumnValuesToBeNull,
    ExpectColumnValuesToBeOfType,
    ExpectColumnValuesToBeUnique,
    ExpectColumnValuesToMatchJsonSchema,
    ExpectColumnValuesToMatchLikePattern,
    ExpectColumnValuesToMatchLikePatternList,
    ExpectColumnValuesToMatchRegex,
    ExpectColumnValuesToMatchRegexList,
    ExpectColumnValuesToMatchStrftimeFormat,
    ExpectColumnValuesToNotBeInSet,
    ExpectColumnValuesToNotBeNull,
    ExpectColumnValuesToNotMatchLikePattern,
    ExpectColumnValuesToNotMatchLikePatternList,
    ExpectColumnValuesToNotMatchRegex,
    ExpectColumnValuesToNotMatchRegexList,
    ExpectColumnValueZScoresToBeLessThan,
    ExpectCompoundColumnsToBeUnique,
    ExpectMulticolumnSumToEqual,
    ExpectMulticolumnValuesToBeUnique,
    ExpectQueryResultsToMatchComparison,
    ExpectSelectColumnValuesToBeUniqueWithinRecord,
    ExpectTableColumnCountToBeBetween,
    ExpectTableColumnCountToEqual,
    ExpectTableColumnsToMatchOrderedList,
    ExpectTableColumnsToMatchSet,
    ExpectTableRowCountToBeBetween,
    ExpectTableRowCountToEqual,
    ExpectTableRowCountToEqualOtherTable,
    UnexpectedRowsExpectation,
)



from typing import Final

import pytest

from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.datasource.fluent.pandas_datasource import (
    ClipboardAsset,
    CSVAsset,
    ExcelAsset,
    FeatherAsset,
    FWFAsset,
    GBQAsset,
    HDFAsset,
    HTMLAsset,
    JSONAsset,
    ORCAsset,
    PandasDatasource,
    ParquetAsset,
    PickleAsset,
    SASAsset,
    SPSSAsset,
    SQLAsset,
    SQLQueryAsset,
    SQLTableAsset,
    StataAsset,
    TableAsset,
    XMLAsset,
    _PandasDataAsset,
)


@pytest.fixture
def datasource(mocker):
    return mocker.Mock(spec=PandasDatasource)


@pytest.fixture
def asset(request, datasource) -> _PandasDataAsset:
    asset = request.param
    asset._datasource = datasource  # same pattern Datasource uses to init Asset
    return asset


PATH_LIKE_STR: Final[str] = "./test/file"
QUERY: Final[str] = "select * from my_table;"
CONNECTION_URL: Final[str] = "connection:url"


def _asset_parameters():
    return [
        pytest.param(
            ClipboardAsset(name="test_asset", type="clipboard"),
            id="ClipboardAsset",
        ),
        pytest.param(
            CSVAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="csv"),
            id="CSVAsset",
        ),
        pytest.param(
            ExcelAsset(name="test_asset", io=PATH_LIKE_STR, type="excel"),
            id="ExcelAsset",
        ),
        pytest.param(
            FeatherAsset(name="test_asset", path=PATH_LIKE_STR, type="feather"),
            id="FeatherAsset",
        ),
        pytest.param(
            FWFAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="fwf"),
            id="FWFAsset",
        ),
        pytest.param(
            GBQAsset(name="test_asset", query=QUERY, type="gbq"),
            id="GBQAsset",
        ),
        pytest.param(
            HDFAsset(name="test_asset", path_or_buf=PATH_LIKE_STR, type="hdf"),
            id="HDFAsset",
        ),
        pytest.param(
            HTMLAsset(name="test_asset", io=PATH_LIKE_STR, type="html"),
            id="HTMLAsset",
        ),
        pytest.param(
            JSONAsset(name="test_asset", path_or_buf=PATH_LIKE_STR, type="json"),
            id="JSONAsset",
        ),
        pytest.param(
            ORCAsset(name="test_asset", path=PATH_LIKE_STR, type="orc"),
            id="ORCAsset",
        ),
        pytest.param(
            ParquetAsset(name="test_asset", path=PATH_LIKE_STR, type="parquet"),
            id="ParquetAsset",
        ),
        pytest.param(
            PickleAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="pickle"),
            id="PickleAsset",
        ),
        pytest.param(
            SQLAsset(name="test_asset", sql=QUERY, con=CONNECTION_URL, type="sql"),
            id="SQLAsset",
        ),
        pytest.param(
            SQLQueryAsset(
                name="test_asset",
                sql=QUERY,
                con=CONNECTION_URL,
                type="sql_query",
            ),
            id="SQLQueryAsset",
        ),
        pytest.param(
            SQLTableAsset(
                name="test_asset",
                table_name="my_table",
                con=CONNECTION_URL,
                type="sql_table",
            ),
            id="SQLTableAsset",
        ),
        pytest.param(
            SASAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="sas"),
            id="SASAsset",
        ),
        pytest.param(
            SPSSAsset(name="test_asset", path=PATH_LIKE_STR, type="spss"),
            id="SPSSAsset",
        ),
        pytest.param(
            StataAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="stata"),
            id="StataAsset",
        ),
        pytest.param(
            TableAsset(name="test_asset", filepath_or_buffer=PATH_LIKE_STR, type="table"),
            id="TableAsset",
        ),
        pytest.param(
            XMLAsset(name="test_asset", path_or_buffer=PATH_LIKE_STR, type="xml"),
            id="XMLAsset",
        ),
    ]


@pytest.mark.unit
@pytest.mark.parametrize("asset", _asset_parameters(), indirect=["asset"])
def test_add_batch_definition_whole_dataframe(asset, datasource):
    # arrange
    name = "batch_def_name"
    expected_batch_definition = BatchDefinition(name=name, partitioner=None)
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_whole_dataframe(name=name)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)



import math
from datetime import datetime, timedelta
from timeit import timeit
from typing import Any

import dateutil
import pytest

from great_expectations.core.suite_parameters import (
    _deduplicate_suite_parameter_dependencies,
    get_suite_parameter_key,
    is_suite_parameter,
    parse_suite_parameter,
)
from great_expectations.exceptions import SuiteParameterError


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, False),
        (1, False),
        ("foo", False),
        ({}, False),
        ({"PARAMETER": 123}, False),
        ({"$PARAMETER": 123}, True),
        ({"$PARAMETER": 123, "another_field": 234}, True),
    ],
)
@pytest.mark.unit
def test_is_suite_parameter(value: Any, expected: bool):
    assert is_suite_parameter(value) == expected


@pytest.mark.unit
def test_get_suite_parameter_key():
    key = "foo"
    assert get_suite_parameter_key({"$PARAMETER": key}) == key


@pytest.mark.unit
def test_parse_suite_parameter():
    # Substitution alone is ok
    assert parse_suite_parameter("a", {"a": 1}) == 1
    assert (
        parse_suite_parameter(
            "my_value",
            {"my_value": 1},
        )
        == 1
    )

    # Very basic arithmetic is allowed as-is:
    assert parse_suite_parameter("1 + 1", {}) == 2

    # So is simple variable substitution:
    assert parse_suite_parameter("a + 1", {"a": 2}) == 3

    # bare decimal is accepted
    assert parse_suite_parameter("my_value * .9", {"my_value": 10}) == 9

    assert parse_suite_parameter("trunc(my_value * 0.9)", {"my_value": 11}) == 9

    # Valid variables but invalid expression is no good
    with pytest.raises(SuiteParameterError) as err:
        parse_suite_parameter("1 / a", {"a": 0})
    assert "Error while evaluating suite parameter expression: division by zero" in str(err.value)

    # It is okay to *substitute* strings in the expression...
    assert parse_suite_parameter("foo", {"foo": "bar"}) == "bar"

    # ...and to have whitespace in substituted values...
    assert parse_suite_parameter("foo", {"foo": "bar "}) == "bar "

    # ...but whitespace is *not* preserved from the parameter name if we evaluate it
    assert parse_suite_parameter("foo ", {"foo": "bar"}) == "bar"  # NOT "bar "

    # We can use multiple parameters...
    assert parse_suite_parameter("foo * bar", {"foo": 2, "bar": 3}) == 6

    # ...but we cannot leave *partially* evaluated expressions (phew!)
    with pytest.raises(SuiteParameterError) as e:
        parse_suite_parameter("foo + bar", {"foo": 2})
    assert "Error while evaluating suite parameter expression: Unknown string format: bar" in str(
        e.value
    )


@pytest.mark.unit
def test_parser_timing():
    """We currently reuse the parser, clearing the stack between calls, which is about 10 times faster than not
    doing so. But these operations are really quick, so this may not be necessary."""  # noqa: E501 # FIXME CoP
    assert (
        timeit(
            "parse_suite_parameter('x', {'x': 1})",
            setup="from great_expectations.core.suite_parameters import parse_suite_parameter",
            number=100,
        )
        < 1
    )


@pytest.mark.unit
def test_math_suite_paramaters():
    assert parse_suite_parameter("sin(2*PI)") == math.sin(math.pi * 2)
    assert parse_suite_parameter("cos(2*PI)") == math.cos(math.pi * 2)
    assert parse_suite_parameter("tan(2*PI)") == math.tan(math.pi * 2)


@pytest.mark.unit
def test_temporal_suite_parameters():
    # allow 1 second for "now" tolerance
    now = datetime.now()  # noqa: DTZ005 # FIXME CoP
    assert (
        (now - timedelta(weeks=1, seconds=3))
        < dateutil.parser.parse(parse_suite_parameter("now() - timedelta(weeks=1, seconds=2)"))
        < now - timedelta(weeks=1, seconds=1)
    )


@pytest.mark.unit
def test_temporal_suite_parameters_complex():
    # allow 1 second for "now" tolerance
    now = datetime.now()  # noqa: DTZ005 # FIXME CoP
    # Choosing "2*3" == 6 weeks shows we can parse an expression inside a kwarg.
    assert (
        (now - timedelta(weeks=2 * 3, seconds=3))
        < dateutil.parser.parse(parse_suite_parameter("now() - timedelta(weeks=2*3, seconds=2)"))
        < now - timedelta(weeks=2 * 3, seconds=1)
    )


@pytest.mark.unit
def test_deduplicate_suite_parameter_dependencies():
    dependencies = {
        "profile": [
            {
                "metric_kwargs_id": {
                    "column=norm": ["expect_column_mean_to_be_between.result.observed_value"]
                }
            },
            {
                "metric_kwargs_id": {
                    "column=norm": ["expect_column_stdev_to_be_between.result.observed_value"]
                }
            },
        ]
    }

    deduplicated = _deduplicate_suite_parameter_dependencies(dependencies)

    # For test, use set to ignore order
    deduplicated["profile"][0]["metric_kwargs_id"]["column=norm"] = set(
        deduplicated["profile"][0]["metric_kwargs_id"]["column=norm"]
    )

    assert deduplicated == {
        "profile": [
            {
                "metric_kwargs_id": {
                    "column=norm": {
                        "expect_column_mean_to_be_between.result.observed_value",
                        "expect_column_stdev_to_be_between.result.observed_value",
                    }
                }
            }
        ]
    }


@pytest.mark.unit
def test_now_suite_parameter():
    """
    now() is unique in the fact that it is the only suite param built-in that has zero arity (takes no arguments).
    The following tests ensure that it is properly parsed and evaluated in a variety of contexts.
    """  # noqa: E501 # FIXME CoP
    # By itself
    res = parse_suite_parameter("now()")
    assert dateutil.parser.parse(res), "Provided suite parameter is not dateutil-parseable"

    # In conjunction with timedelta
    res = parse_suite_parameter("now() - timedelta(weeks=1)")
    assert dateutil.parser.parse(res), "Provided suite parameter is not dateutil-parseable"

    # Require parens to actually invoke
    with pytest.raises(SuiteParameterError):
        parse_suite_parameter("now")



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

COL_WITHOUT_TIE = "without_tie"
COL_WITH_TIE = "with_tie"
STRING_COL = "string_col"

A_LOT_OF_NONES = [None] * 10
DATA = pd.DataFrame(
    {
        COL_WITHOUT_TIE: [0, 0, 1, 1, 2, 2, 3, 4, 4, 4, *A_LOT_OF_NONES],
        COL_WITH_TIE: [0, 1, 1, 2, 2, 2, 3, 4, 4, 4, *A_LOT_OF_NONES],
    },
    dtype="object",
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMostCommonValueToBeInSet(column=COL_WITHOUT_TIE, value_set=[4])
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": [4]}


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({COL_WITHOUT_TIE: ["foo", "foo", "foo", "bar", "baz"]}),
)
def test_strings(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMostCommonValueToBeInSet(
        column=COL_WITHOUT_TIE, value_set=["foo"]
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMostCommonValueToBeInSet(
                column=COL_WITH_TIE, value_set=[2, 4, 100], ties_okay=True
            ),
            id="tie_and_value_set_has_extra_values",
        ),
        pytest.param(
            # This test case surprises me
            gxe.ExpectColumnMostCommonValueToBeInSet(
                column=COL_WITH_TIE, value_set=[2], ties_okay=True
            ),
            id="value_set_is_subset",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnMostCommonValueToBeInSet
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMostCommonValueToBeInSet(
                column=COL_WITH_TIE, value_set=[100, 101], ties_okay=True
            ),
            id="no_matches_to_value_set",
        ),
        pytest.param(
            # This feels like it conflicts with the 2 positive cases above
            gxe.ExpectColumnMostCommonValueToBeInSet(
                column=COL_WITH_TIE, value_set=[2, 100], ties_okay=False
            ),
            id="value_set_and_observed_have_partial_overlap",
        ),
        pytest.param(
            gxe.ExpectColumnMostCommonValueToBeInSet(
                column=COL_WITH_TIE, value_set=[2, 4], ties_okay=False
            ),
            id="ties_not_okay",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnMostCommonValueToBeInSet
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_ties_okay_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_most_common_value_to_be_in_set"
    expectation = gxe.ExpectColumnMostCommonValueToBeInSet(
        column=COL_WITH_TIE,
        value_set=[2, 4, 100],
        ties_okay={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

import copy
import logging
import re
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Generator, List, Optional, Type

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_spec import PathBatchSpec, S3BatchSpec
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    FilePathDataConnector,
    MissingFilePathTemplateMapFnError,
    sanitize_prefix_for_gcs_and_s3,
)

if TYPE_CHECKING:
    from botocore.client import BaseClient

    from great_expectations.alias_types import PathStr
    from great_expectations.core.batch import LegacyBatchDefinition


logger = logging.getLogger(__name__)


class _S3Options(pydantic.BaseModel):
    s3_prefix: str = ""
    s3_delimiter: str = "/"
    s3_max_keys: int = 1000
    s3_recursive_file_discovery: bool = False


class S3DataConnector(FilePathDataConnector):
    """Extension of FilePathDataConnector used to connect to S3.


    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
        s3_client: Reference to instantiated AWS S3 client handle
        bucket (str): bucket for S3
        prefix (str): S3 prefix
        delimiter (str): S3 delimiter
        max_keys (int): S3 max_keys (default is 1000)
        recursive_file_discovery (bool): Flag to indicate if files should be searched recursively from subfolders
        file_path_template_map_fn: Format function mapping path to fully-qualified resource on S3
        whole_directory_path_override: If present, treat entire directory as single Asset
    """  # noqa: E501 # FIXME CoP

    asset_level_option_keys: ClassVar[tuple[str, ...]] = (
        "s3_prefix",
        "s3_delimiter",
        "s3_max_keys",
        "s3_recursive_file_discovery",
    )
    asset_options_type: ClassVar[Type[_S3Options]] = _S3Options

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_asset_name: str,
        s3_client: BaseClient,
        bucket: str,
        prefix: str = "",
        delimiter: str = "/",
        max_keys: int = 1000,
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        self._s3_client: BaseClient = s3_client

        self._bucket: str = bucket

        self._prefix: str = prefix
        self._sanitized_prefix: str = sanitize_prefix_for_gcs_and_s3(text=prefix)

        self._delimiter: str = delimiter
        self._max_keys: int = max_keys

        self._recursive_file_discovery = recursive_file_discovery

        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_data_connector(  # noqa: PLR0913 # FIXME CoP
        cls,
        datasource_name: str,
        data_asset_name: str,
        s3_client: BaseClient,
        bucket: str,
        prefix: str = "",
        delimiter: str = "/",
        max_keys: int = 1000,
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> S3DataConnector:
        """Builds "S3DataConnector", which links named DataAsset to AWS S3.

        Args:
            datasource_name: The name of the Datasource associated with this "S3DataConnector" instance
            data_asset_name: The name of the DataAsset using this "S3DataConnector" instance
            s3_client: S3 Client reference handle
            bucket: bucket for S3
            prefix: S3 prefix
            delimiter: S3 delimiter
            max_keys: S3 max_keys (default is 1000)
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders
            file_path_template_map_fn: Format function mapping path to fully-qualified resource on S3
            whole_directory_path_override: If present, treat entire directory as single Asset

        Returns:
            Instantiated "S3DataConnector" object
        """  # noqa: E501 # FIXME CoP
        return S3DataConnector(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            s3_client=s3_client,
            bucket=bucket,
            prefix=prefix,
            delimiter=delimiter,
            max_keys=max_keys,
            recursive_file_discovery=recursive_file_discovery,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_test_connection_error_message(
        cls,
        data_asset_name: str,
        bucket: str,
        prefix: str = "",
        delimiter: str = "/",
        recursive_file_discovery: bool = False,
    ) -> str:
        """Builds helpful error message for reporting issues when linking named DataAsset to Microsoft Azure Blob Storage.

        Args:
            data_asset_name: The name of the DataAsset using this "AzureBlobStorageDataConnector" instance
            bucket: bucket for S3
            prefix: S3 prefix
            delimiter: S3 delimiter
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders

        Returns:
            Customized error message
        """  # noqa: E501 # FIXME CoP
        test_connection_error_message_template: str = 'No file in bucket "{bucket}" with prefix "{prefix}" and recursive file discovery set to "{recursive_file_discovery}" found using delimiter "{delimiter}" for DataAsset "{data_asset_name}".'  # noqa: E501 # FIXME CoP
        return test_connection_error_message_template.format(
            **{
                "data_asset_name": data_asset_name,
                "bucket": bucket,
                "prefix": prefix,
                "delimiter": delimiter,
                "recursive_file_discovery": recursive_file_discovery,
            }
        )

    @override
    def build_batch_spec(self, batch_definition: LegacyBatchDefinition) -> S3BatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (LegacyBatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec: PathBatchSpec = super().build_batch_spec(batch_definition=batch_definition)
        return S3BatchSpec(batch_spec)

    # Interface Method
    @override
    def get_data_references(self) -> List[str]:
        query_options: dict = {
            "Bucket": self._bucket,
            "Prefix": self._sanitized_prefix,
            "Delimiter": self._delimiter,
            "MaxKeys": self._max_keys,
        }
        path_list: List[str] = list(
            list_s3_keys(
                s3=self._s3_client,
                query_options=query_options,
                iterator_dict={},
                recursive=self._recursive_file_discovery,
            )
        )
        return path_list

    # Interface Method
    @override
    def _get_full_file_path(self, path: str) -> str:
        # If the path is already a fully qualified S3 URL (starts with s3://), return it as-is
        # This handles the case of whole_directory_path_override which is already fully qualified
        if path.startswith("s3://"):
            return path

        if self._file_path_template_map_fn is None:
            raise MissingFilePathTemplateMapFnError()

        template_arguments = {
            "bucket": self._bucket,
            "path": path,
        }

        return self._file_path_template_map_fn(**template_arguments)

    @override
    def _preprocess_batching_regex(self, regex: re.Pattern) -> re.Pattern:
        regex = re.compile(f"{re.escape(self._sanitized_prefix)}{regex.pattern}")
        return super()._preprocess_batching_regex(regex=regex)


def list_s3_keys(  # noqa: C901 #  too complex
    s3, query_options: dict, iterator_dict: dict, recursive: bool = False
) -> Generator[str, None, None]:
    """
    For InferredAssetS3DataConnector, we take bucket and prefix and search for files using RegEx at and below the level
    specified by that bucket and prefix.  However, for ConfiguredAssetS3DataConnector, we take bucket and prefix and
    search for files using RegEx only at the level specified by that bucket and prefix.  This restriction for the
    ConfiguredAssetS3DataConnector is needed, because paths on S3 are comprised not only the leaf file name but the
    full path that includes both the prefix and the file name.  Otherwise, in the situations where multiple data assets
    share levels of a directory tree, matching files to data assets will not be possible, due to the path ambiguity.
    :param s3: s3 client connection
    :param query_options: s3 query attributes ("Bucket", "Prefix", "Delimiter", "MaxKeys")
    :param iterator_dict: dictionary to manage "NextContinuationToken" (if "IsTruncated" is returned from S3)
    :param recursive: True for InferredAssetS3DataConnector and False for ConfiguredAssetS3DataConnector (see above)
    :return: string valued key representing file path on S3 (full prefix and leaf file name)
    """  # noqa: E501 # FIXME CoP
    if iterator_dict is None:
        iterator_dict = {}

    if "continuation_token" in iterator_dict:
        query_options.update({"ContinuationToken": iterator_dict["continuation_token"]})

    logger.debug(f"Fetching objects from S3 with query options: {query_options}")

    s3_objects_info: dict = s3.list_objects_v2(**query_options)

    if not any(key in s3_objects_info for key in ["Contents", "CommonPrefixes"]):
        raise ValueError("S3 query may not have been configured correctly.")  # noqa: TRY003 # FIXME CoP

    if "Contents" in s3_objects_info:
        keys: List[str] = [item["Key"] for item in s3_objects_info["Contents"] if item["Size"] > 0]
        yield from keys

    if recursive and "CommonPrefixes" in s3_objects_info:
        common_prefixes: List[Dict[str, Any]] = s3_objects_info["CommonPrefixes"]
        for prefix_info in common_prefixes:
            query_options_tmp: dict = copy.deepcopy(query_options)
            query_options_tmp.update({"Prefix": prefix_info["Prefix"]})
            # Recursively fetch from updated prefix
            yield from list_s3_keys(
                s3=s3,
                query_options=query_options_tmp,
                iterator_dict={},
                recursive=recursive,
            )

    if s3_objects_info["IsTruncated"]:
        iterator_dict["continuation_token"] = s3_objects_info["NextContinuationToken"]
        # Recursively fetch more
        yield from list_s3_keys(
            s3=s3,
            query_options=query_options,
            iterator_dict=iterator_dict,
            recursive=recursive,
        )

    if "continuation_token" in iterator_dict:
        # Make sure we clear the token once we've gotten fully through
        del iterator_dict["continuation_token"]



from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnNullCountResult(MetricResult[int]): ...


class ColumnNullCount(ColumnMetric[ColumnNullCountResult]):
    """Count of null values in a column"""

    name = "column_values.null.count"



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

BOTO_NOT_IMPORTED = NotImported(
    "AWS S3 connection components are not installed, please 'pip install boto3 botocore'"
)
REDSHIFT_NOT_IMPORTED = NotImported(
    "AWS Redshift connection component is not installed, please 'pip install sqlalchemy_redshift'"
)
ATHENA_NOT_IMPORTED = NotImported(
    "AWS Athena connection component is not installed, please 'pip install pyathena[SQLAlchemy]>=2.0.0,<3'"  # noqa: E501 # FIXME CoP
)

try:
    import boto3
except ImportError:
    boto3 = BOTO_NOT_IMPORTED

try:
    import botocore
except ImportError:
    botocore = BOTO_NOT_IMPORTED

try:
    from botocore.client import Config
except ImportError:
    Config = BOTO_NOT_IMPORTED

try:
    from botocore import exceptions
except ImportError:
    exceptions = BOTO_NOT_IMPORTED

try:
    import sqlalchemy_redshift
except ImportError:
    sqlalchemy_redshift = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift import dialect as redshiftdialect
except (ImportError, AttributeError):
    redshiftdialect = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import CHAR
except (ImportError, AttributeError):
    CHAR = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import VARCHAR
except (ImportError, AttributeError):
    VARCHAR = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import INTEGER
except (ImportError, AttributeError):
    INTEGER = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import SMALLINT
except (ImportError, AttributeError):
    SMALLINT = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import BIGINT
except (ImportError, AttributeError):
    BIGINT = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import TIMESTAMP
except (ImportError, AttributeError):
    TIMESTAMP = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import DATE
except (ImportError, AttributeError):
    DATE = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import DOUBLE_PRECISION
except (ImportError, AttributeError):
    DOUBLE_PRECISION = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import BOOLEAN
except (ImportError, AttributeError):
    BOOLEAN = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import DECIMAL
except (ImportError, AttributeError):
    DECIMAL = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import GEOMETRY
except (ImportError, AttributeError):
    GEOMETRY = REDSHIFT_NOT_IMPORTED

try:
    from sqlalchemy_redshift.dialect import SUPER
except (ImportError, AttributeError):
    SUPER = REDSHIFT_NOT_IMPORTED

try:
    import pyathena  # type: ignore[import-not-found] # FIXME CoP
except ImportError:
    pyathena = ATHENA_NOT_IMPORTED

try:
    from pyathena import sqlalchemy_athena
except (ImportError, AttributeError):
    sqlalchemy_athena = ATHENA_NOT_IMPORTED

try:
    from pyathena.sqlalchemy_athena import (  # type: ignore[import-not-found] # FIXME CoP
        types as athenatypes,
    )
except (ImportError, AttributeError):
    athenatypes = ATHENA_NOT_IMPORTED


class REDSHIFT_TYPES:
    """Namespace for Redshift dialect types."""

    CHAR = CHAR
    VARCHAR = VARCHAR
    INTEGER = INTEGER
    SMALLINT = SMALLINT
    BIGINT = BIGINT
    TIMESTAMP = TIMESTAMP
    DATE = DATE
    DOUBLE_PRECISION = DOUBLE_PRECISION
    BOOLEAN = BOOLEAN
    DECIMAL = DECIMAL
    GEOMETRY = GEOMETRY
    SUPER = SUPER



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar, List, Optional
from uuid import UUID

from great_expectations import __version__ as gx_version
from great_expectations.analytics.config import get_config
from great_expectations.compatibility.typing_extensions import override


@dataclass(frozen=True)
class Action:
    """Action is the "verb" representing what happened.

    Attributes:
        name: A description of what happened. For example (<object>.<verb>) "validation_result.saved" or "token.deleted"
    """  # noqa: E501 # FIXME CoP

    name: str

    @override
    def __repr__(self):
        return self.name


@dataclass
class Event:
    """Details of an analytics event.

    Attributes:
        action: The "verb" describing what action this event represents.
    """

    action: Action

    @property
    def data_context_id(self) -> UUID | None:
        return get_config().data_context_id

    @property
    def organization_id(self) -> UUID | None:
        return get_config().organization_id

    @property
    def oss_id(self) -> UUID | None:
        return get_config().oss_id

    @property
    def user_id(self) -> UUID | None:
        return get_config().user_id

    @property
    def user_agent_str(self) -> str | None:
        return get_config().user_agent_str

    @property
    def mode(self) -> str | None:
        return get_config().mode

    @property
    def distinct_id(self) -> UUID | None:
        """The distinct_id is the primary key for identifying
        analytics events. It is the user_id if it is set
        (e.g. in a Cloud context), otherwise the oss_id.
        """
        return self.user_id or self.oss_id

    _allowed_actions: ClassVar[Optional[List[Action]]] = None

    def __post_init__(self):
        allowed_actions = self.get_allowed_actions()
        if allowed_actions is not None and self.action not in self.get_allowed_actions():
            error_msg = f"Action [{self.action}] must be one of {self.get_allowed_actions()}"
            raise ValueError(error_msg)  # FIXME CoP

    @classmethod
    def get_allowed_actions(cls):
        return cls._allowed_actions

    def properties(self) -> dict:
        props = {
            "data_context_id": self.data_context_id,
            "oss_id": self.oss_id,
            "gx_version": gx_version,
            "service": "gx-core",
            "user_agent_str": self.user_agent_str,
            "mode": self.mode,
            # https://posthog.com/docs/libraries/python#person-profiles-and-properties
            "$process_person_profile": not get_config().remove_profile,
        }

        if self.user_id is not None:
            props.update({"user_id": self.user_id, "organization_id": self.organization_id})

        return {**props, **self._properties()}

    def _properties(self) -> dict:
        """Returns event specific properties.

        Events can be extended with additional properties. Subclasses should define
        a _properties method to return them all as a dict.

        Returns:
            A dict representation of the event specific properties
        """
        return {}



from __future__ import annotations

import json
import re
from typing import Any, Final

import pytest

from great_expectations.core.batch_definition import PartitionerT
from great_expectations.core.partitioners import (
    ColumnPartitioner,
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    FileNamePartitionerDaily,
    FileNamePartitionerMonthly,
    FileNamePartitionerYearly,
)
from great_expectations.datasource.fluent import BatchRequest


@pytest.mark.unit
@pytest.mark.parametrize(
    "optional_batch_request_config,parsed_batch_slice",
    [
        pytest.param(
            {
                "batch_slice": "[4:20:2]",
            },
            slice(4, 20, 2),
            id="no options, batch_slice: str",
        ),
        pytest.param(
            {
                "batch_slice": (1, 27, 3),
            },
            slice(1, 27, 3),
            id="no options, batch_slice: tuple",
        ),
        pytest.param(
            {
                "options": {"season": "summer"},
            },
            slice(0, None, None),
            id="no batch_slice, options: str",
        ),
        pytest.param(
            {
                "options": {"month": 7},
            },
            slice(0, None, None),
            id="no batch_slice, options: int",
        ),
        pytest.param(
            {
                "options": {"greeting": "hello"},
                "batch_slice": -1,
            },
            slice(-1, None, None),
            id="options: str, batch_slice: neg int",
        ),
        pytest.param(
            {
                "options": {"month": None},
                "batch_slice": "[1:4]",
            },
            slice(1, 4, None),
            id="options: None, batch_slice: str",
        ),
    ],
)
def test_batch_request_config_serialization_round_trips(
    optional_batch_request_config: dict, parsed_batch_slice: slice
) -> None:
    datasource_name: Final[str] = "my_datasource"
    data_asset_name: Final[str] = "my_data_asset"
    batch_request_config: dict[str, Any] = {
        "datasource_name": datasource_name,
        "data_asset_name": data_asset_name,
        "partitioner": ColumnPartitionerYearly(column_name="my_column"),
    }
    batch_request_config.update(optional_batch_request_config)
    batch_request = BatchRequest[ColumnPartitioner](**batch_request_config)
    assert batch_request.datasource_name == datasource_name
    assert batch_request.data_asset_name == data_asset_name
    # options is optional and an empty dict by default
    assert batch_request.options == batch_request_config.get("options", {})
    # BatchRequest always has a slice associated with it,
    # even if it is slice(0, None, None) (all elements in sequence).
    assert batch_request.batch_slice == parsed_batch_slice

    #
    # dict
    #
    batch_request_dict = batch_request.dict()
    assert batch_request_dict["datasource_name"] == datasource_name
    assert batch_request_dict["data_asset_name"] == data_asset_name
    # options is optional and an empty dict by default
    assert batch_request_dict["options"] == batch_request_config.get("options", {})
    # Even though BatchRequest.batch_slice has a slice object,
    # it will serialize to None since batch_slice wasn't passed in.
    assert batch_request_dict["batch_slice"] == batch_request_config.get("batch_slice", None)

    #
    # json
    #
    batch_request_json = batch_request.json()

    # convert options dict to json string
    options_json = json.dumps(batch_request_config.get("options", {}))

    # convert batch_slice value to json string
    batch_slice_json = json.dumps(batch_request_config.get("batch_slice", None))

    assert batch_request_json == (
        "{"
        f'"datasource_name": "{datasource_name}", '
        f'"data_asset_name": "{data_asset_name}", '
        f'"options": {options_json}, '
        '"partitioner": {'
        '"column_name": "my_column", '
        '"sort_ascending": true, '
        '"method_name": "partition_on_year"'
        "}, "
        f'"batch_slice": {batch_slice_json}'
        "}"
    )


def _partitioner_test_cases():
    return [
        pytest.param(
            ColumnPartitionerDaily(
                column_name="foo",
                sort_ascending=False,
            ),
            id="Sql Daily",
        ),
        pytest.param(
            ColumnPartitionerMonthly(
                column_name="foo",
                sort_ascending=False,
            ),
            id="Sql Monthly",
        ),
        pytest.param(
            ColumnPartitionerYearly(
                column_name="foo",
                sort_ascending=False,
            ),
            id="Sql Yearly",
        ),
        pytest.param(
            FileNamePartitionerDaily(
                regex=re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2}).csv"),
                sort_ascending=False,
            ),
            id="Regex Daily",
        ),
        pytest.param(
            FileNamePartitionerMonthly(
                regex=re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv"),
                sort_ascending=False,
            ),
            id="Regex Monthly",
        ),
        pytest.param(
            FileNamePartitionerYearly(
                regex=re.compile(r"data_(?P<year>\d{4}).csv"),
                sort_ascending=False,
            ),
            id="Regex Yearly",
        ),
        pytest.param(
            None,
            id="None",
        ),
    ]


@pytest.mark.unit
@pytest.mark.parametrize(
    "partitioner",
    _partitioner_test_cases(),
)
def test_batch_request_config_partitioner_dict_round_trip_serialization(
    partitioner: PartitionerT,
) -> None:
    # Using Any here effectively turns off type checking. This test demonstrates that the
    # BatchRequest doesn't need to be parameterized for Pydantic to correctly deserialize from dict
    batch_request = BatchRequest[Any](
        datasource_name="test-datasource", data_asset_name="test-asset", partitioner=partitioner
    )

    batch_request_dict = batch_request.dict()
    assert BatchRequest(**batch_request_dict) == batch_request


@pytest.mark.unit
@pytest.mark.parametrize(
    "partitioner",
    _partitioner_test_cases(),
)
def test_batch_request_config_partitioner_json_round_trip_serialization(
    partitioner: PartitionerT,
) -> None:
    # Using Any here effectively turns off type checking. This test demonstrates that the
    # BatchRequest doesn't need to be parameterized for Pydantic to correctly deserialize from JSON
    batch_request = BatchRequest[Any](
        datasource_name="test-datasource", data_asset_name="test-asset", partitioner=partitioner
    )

    batch_request_json = batch_request.json()
    assert BatchRequest.parse_raw(batch_request_json) == batch_request
